import numpy as np


def mahalanobis(X1, X2, S1, S2):
    r"""Calculate Mahalanobis distance between two points with covariance matrices.
    
    Arguments:
        X1: First point coordinates
        X2: Second point coordinates
        S1: Covariance matrix for first point
        S2: Covariance matrix for second point
        
    Returns:
        float: Mahalanobis distance
    """
    S_inv = np.linalg.inv(S1 + S2)
    diff = (X1 - X2).reshape(-1, 1)
    return np.matmul(np.matmul(diff.T, S_inv), diff)

def isint(x):
    if isinstance(x, int):
        return True
    if isinstance(x, str):
        return False
    try:
        a = float(x)
        b = int(a)
    except ValueError:
        return False
    else:
        return a == b

def isstr(x):
    return isinstance(x, str)

def scale_to_range(x, a=0, b=1):
    r"""Scale array values to specified range.
    
    Arguments:
        x: Input array to scale
        a (float): Lower bound of target range (default: 0)
        b (float): Upper bound of target range (default: 1)
        
    Returns:
        np.ndarray: Scaled array
    """
    return ((x - x.min()) / (x.max() - x.min())) * (b - a) + a

class Lineage:
    r"""Represents a lineage trajectory through clusters.
    
    Arguments:
        clusters: List of cluster IDs forming the lineage
    """
    def __init__(self, clusters):
        self.clusters = clusters

    def __len__(self):
        return len(self.clusters)

    def __repr__(self):
        return 'Lineage' + str(self.clusters)

    def __iter__(self):
        for c in self.clusters:
            yield c

import seaborn as sns

from matplotlib import pyplot as plt
from matplotlib.patches import Patch
from mpl_toolkits.axes_grid1 import make_axes_locatable

import math
import numpy as np
from matplotlib.colors import ListedColormap
from matplotlib.cm import hsv


def generate_colormap(number_of_distinct_colors: int = 80):
    if number_of_distinct_colors == 0:
        number_of_distinct_colors = 80

    number_of_shades = 7
    number_of_distinct_colors_with_multiply_of_shades = int(math.ceil(number_of_distinct_colors / number_of_shades) * number_of_shades)

    # Create an array with uniformly drawn floats taken from <0, 1) partition
    linearly_distributed_nums = np.arange(number_of_distinct_colors_with_multiply_of_shades) / number_of_distinct_colors_with_multiply_of_shades

    # We are going to reorganise monotonically growing numbers in such way that there will be single array with saw-like pattern
    #     but each saw tooth is slightly higher than the one before
    # First divide linearly_distributed_nums into number_of_shades sub-arrays containing linearly distributed numbers
    arr_by_shade_rows = linearly_distributed_nums.reshape(number_of_shades, number_of_distinct_colors_with_multiply_of_shades // number_of_shades)

    # Transpose the above matrix (columns become rows) - as a result each row contains saw tooth with values slightly higher than row above
    arr_by_shade_columns = arr_by_shade_rows.T

    # Keep number of saw teeth for later
    number_of_partitions = arr_by_shade_columns.shape[0]

    # Flatten the above matrix - join each row into single array
    nums_distributed_like_rising_saw = arr_by_shade_columns.reshape(-1)

    # HSV colour map is cyclic (https://matplotlib.org/tutorials/colors/colormaps.html#cyclic), we'll use this property
    initial_cm = hsv(nums_distributed_like_rising_saw)

    lower_partitions_half = number_of_partitions // 2
    upper_partitions_half = number_of_partitions - lower_partitions_half

    # Modify lower half in such way that colours towards beginning of partition are darker
    # First colours are affected more, colours closer to the middle are affected less
    lower_half = lower_partitions_half * number_of_shades
    for i in range(3):
        step_size = 0.8
        if lower_half != 0:
            step_size /= lower_half

        initial_cm[0:lower_half, i] *= np.arange(0.2, 1, step_size)        

    # Modify second half in such way that colours towards end of partition are less intense and brighter
    # Colours closer to the middle are affected less, colours closer to the end are affected more
    for i in range(3):
        for j in range(upper_partitions_half):
            modifier = np.ones(number_of_shades) - initial_cm[lower_half + j * number_of_shades: lower_half + (j + 1) * number_of_shades, i]
            modifier = j * modifier / upper_partitions_half
            initial_cm[lower_half + j * number_of_shades: lower_half + (j + 1) * number_of_shades, i] += modifier

    return ListedColormap(initial_cm)

class SlingshotPlotter:
    def __init__(self, sling):
        self.sling = sling

    def clusters(self, ax, labels=None, s=8, alpha=1., color_mode='clusters'):
        fig = plt.gcf()
        sling = self.sling
        if labels is None:
            labels = np.arange(sling.num_clusters)

        # Plot clusters and start cluster
        ax.scatter(
            sling.cluster_centres[sling.start_node][0],
            sling.cluster_centres[sling.start_node][1], c='red')

        if color_mode == 'clusters':
            colors = np.array(sns.color_palette('cubehelix', n_colors=sling.num_clusters))
            colors = generate_colormap(sling.num_clusters)

            handles = [
                Patch(color=colors.colors[k], label=labels[k]) for k in range(sling.num_clusters)
            ]
            ax.legend(handles=handles)
            colors = colors.colors[sling.cluster_label_indices]
        elif color_mode == 'pseudotime':
            colors = np.zeros_like(self.sling.curves[0].pseudotimes_interp)
            for l_idx, lineage in enumerate(sling.lineages):
                curve = self.sling.curves[l_idx]
                cell_mask = np.logical_or.reduce(
                    np.array([sling.cluster_label_indices == k for k in lineage]))
                colors[cell_mask] = curve.pseudotimes_interp[cell_mask]
        elif type(color_mode) is np.array:
            colors = color_mode
        else:
            colors = 'black'

        main_scatter = ax.scatter(sling.data[:, 0], sling.data[:, 1],
                   c=colors,
                   s=s,
                   alpha=alpha)

        if color_mode == 'pseudotime':
            divider = make_axes_locatable(ax)
            cax = divider.append_axes('right', size='5%', pad=0.05)
            fig.colorbar(main_scatter, cax=cax, orientation='vertical')

    def curves(self, ax, curves):
        for l_idx, curve in enumerate(curves):
            s_interp, p_interp, order = curve.unpack_params()
            ax.plot(
                p_interp[order, 0],
                p_interp[order, 1],
                label=f'Lineage {l_idx}',
                alpha=1)
            ax.legend()

    def network(self, cluster_to_label, figsize=(8, 10)):
        import networkx as nx
        from networkx.drawing.nx_agraph import graphviz_layout
        plt.figure(figsize=figsize)
        G = nx.DiGraph(scale=0.02)
        lineages = self.sling.lineages
        root = cluster_to_label[lineages[0].clusters[0]]

        for lineage in lineages:
            parent = root
            for l in lineage:
                node = cluster_to_label[l]
                G.add_node(node)
                G.add_edge(parent, node)
                parent = node

        plt.title('Lineages')
        pos = graphviz_layout(G, prog='dot')
        label_options = dict(
            ec="k", fc='b', alpha=0.9,
            boxstyle='round,pad=0.2'
        )

        nx.draw(
            G, pos,
            arrows=True,
            node_size=[len(v) * 100 for v in G.nodes()]
        )
        nx.draw_networkx_labels(G, pos, font_size=14, font_color='w', bbox=label_options)


from typing import Union

import numpy as np
from anndata import AnnData


from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.interpolate import interp1d
from sklearn.neighbors import KernelDensity
from collections import deque
from tqdm.autonotebook import tqdm



class Slingshot:
    def __init__(
            self,
            data: Union[AnnData, np.ndarray],
            cluster_labels_onehot=None,
            celltype_key=None,
            obsm_key='X_umap',
            start_node=0,
            end_nodes=None,
            debug_level=None
    ):
        r"""Initialize Slingshot trajectory inference object.
        
        Arguments:
            data: AnnData object or numpy array with dimensionality-reduced data (num_cells, 2)
            cluster_labels_onehot: One-hot encoded cluster labels (default: None)
            celltype_key (str): Key in AnnData.obs for cell type labels (default: None)
            obsm_key (str): Key in AnnData.obsm for dimensionality-reduced data (default: 'X_umap')
            start_node: Starting cluster for trajectory (default: 0)
            end_nodes: List of terminal clusters (default: None)
            debug_level: Debug verbosity level (default: None)
        """
        if isinstance(data, AnnData):
            assert celltype_key is not None, "Must provide celltype key if data is an AnnData object"
            cluster_labels = data.obs[celltype_key]

            if isint(cluster_labels[0]):
                cluster_max = cluster_labels.max()
                self.cluster_label_indices = cluster_labels
            elif isstr(cluster_labels[0]):
                cluster_max = len(np.unique(cluster_labels))
                # Convert list of str labels into a list of int indices
                self.cluster_label_indices = np.array([np.where(np.unique(cluster_labels) == label)[0][0] for label in cluster_labels])
            else:
                raise ValueError("Unexpected cluster label dtype.")
            self.cluster_dict=dict(zip(cluster_labels,self.cluster_label_indices))
            cluster_labels_onehot = np.zeros((cluster_labels.shape[0], cluster_max + 1))
            cluster_labels_onehot[np.arange(cluster_labels.shape[0]), self.cluster_label_indices] = 1

            data = data.obsm[obsm_key]
        else:
            assert cluster_labels_onehot is not None, "Must provide cluster labels if data is not an AnnData object"
            cluster_labels = self.cluster_labels_onehot.argmax(axis=1)
        self.data = data
        self.cluster_labels_onehot = cluster_labels_onehot
        self.cluster_labels = cluster_labels
        self.num_clusters = self.cluster_label_indices.max() + 1
        self.start_node = self.cluster_dict[start_node]
        #self.end_nodes = [] if end_nodes is None else end_nodes
        self.end_nodes = [] if end_nodes is None else [self.cluster_dict[end_node] for end_node in end_nodes]
        cluster_centres = [data[self.cluster_label_indices == k].mean(axis=0) for k in range(self.num_clusters)]
        self.cluster_centres = np.stack(cluster_centres)
        self.lineages = None      # list of Lineages
        self.cluster_lineages = None # lineages belonging to each cluster
        self.curves = None   # list of principle curves len = #lineages
        self.cell_weights = None  # weights indicating cluster assignments
        self.distances = None
        self.branch_clusters = None
        self._tree = None

        # Plotting and printing
        debug_level = 0 if debug_level is None else dict(verbose=1)[debug_level]
        self.debug_level = debug_level
        self._set_debug_axes(None)
        self.plotter = SlingshotPlotter(self)

        # Construct smoothing kernel for the shrinking step
        self.kernel_x = np.linspace(-3, 3, 512)
        kde = KernelDensity(bandwidth=1., kernel='gaussian')
        kde.fit(np.zeros((self.kernel_x.shape[0], 1)))
        self.kernel_y = np.exp(kde.score_samples(self.kernel_x.reshape(-1, 1)))

    @property
    def tree(self):
        if self._tree is None:
            self.construct_mst(self.start_node)
        return self._tree

    def load_params(self, filepath):
        if self.curves is None:
            self.get_lineages()
        params = np.load(filepath, allow_pickle=True).item()
        self.curves = params['curves']   # list of principle curves len = #lineages
        self.cell_weights = params['cell_weights']  # weights indicating cluster assignments
        self.distances = params['distances']

    def save_params(self, filepath):
        params = dict(
            curves=self.curves,
            cell_weights=self.cell_weights,
            distances=self.distances
        )
        np.save(filepath, params)

    def _set_debug_axes(self, axes):
        self.debug_axes = axes
        self.debug_plot_mst = axes is not None
        self.debug_plot_lineages = axes is not None
        self.debug_plot_avg = axes is not None

    def construct_mst(self, start_node):
        r"""Construct minimum spanning tree for trajectory inference.
        
        Arguments:
            start_node: Starting cluster node for the MST
            
        Returns:
            dict: Dictionary mapping clusters to their children in the MST
        """
        # Calculate empirical covariance of clusters
        emp_covs = np.stack([np.cov(self.data[self.cluster_label_indices == i].T) for i in range(self.num_clusters)])
        dists = np.zeros((self.num_clusters, self.num_clusters))
        for i in range(self.num_clusters):
            for j in range(i, self.num_clusters):
                dist = mahalanobis(
                    self.cluster_centres[i],
                    self.cluster_centres[j],
                    emp_covs[i],
                    emp_covs[j]
                )
                dists[i, j] = dist
                dists[j, i] = dist

        # Find minimum spanning tree excluding end nodes
        mst_dists = np.delete(np.delete(dists, self.end_nodes, axis=0), self.end_nodes, axis=1)  # Delete end nodes
        tree = minimum_spanning_tree(mst_dists)
        # On the left: indices with ends removed; on the right: index into an array where the ends are skipped
        index_mapping = np.array([c for c in range(self.num_clusters - len(self.end_nodes))])
        for i, end_node in enumerate(self.end_nodes):
            index_mapping[end_node - i:] += 1

        connections = {k: list() for k in range(self.num_clusters)}
        cx = tree.tocoo()
        for i, j, v in zip(cx.row, cx.col, cx.data):
            i = index_mapping[i]
            j = index_mapping[j]
            connections[i].append(j)
            connections[j].append(i)

        for end in self.end_nodes:
            i = np.argmin(np.delete(dists[end], self.end_nodes))
            connections[i].append(end)
            connections[end].append(i)

        # for i,j,v in zip(cx.row, cx.col, cx.data):
        visited = [False for _ in range(self.num_clusters)]
        queue = list()
        queue.append(start_node)
        children = {k: list() for k in range(self.num_clusters)}
        while len(queue) > 0: # BFS to construct children dict
            current_node = queue.pop()
            visited[current_node] = True
            for child in connections[current_node]:
                if not visited[child]:
                    children[current_node].append(child)
                    queue.append(child)

        # Plot clusters and MST
        if self.debug_plot_mst:
            self.plotter.clusters(self.debug_axes[0, 0], alpha=0.5)
            for root, kids in children.items():
                for child in kids:
                    start = [self.cluster_centres[root][0], self.cluster_centres[child][0]]
                    end = [self.cluster_centres[root][1], self.cluster_centres[child][1]]
                    self.debug_axes[0, 0].plot(start, end, c='black')
            self.debug_plot_mst = False

        self._tree = children
        return children

    def fit(self, num_epochs=10, debug_axes=None):
        self._set_debug_axes(debug_axes)
        if self.curves is None:  # Initial curves and pseudotimes:
            self.get_lineages()
            self.construct_initial_curves()
            self.cell_weights = [self.cluster_labels_onehot[:, self.lineages[l].clusters].sum(axis=1)
                                 for l in range(len(self.lineages))]
            self.cell_weights = np.stack(self.cell_weights, axis=1)

        for epoch in tqdm(range(num_epochs)):
            # Calculate cell weights
            # cell weight is a matrix #cells x #lineages indicating cell-lineage assignment
            self.calculate_cell_weights()

            # Fit principal curve for all lineages using existing curves
            self.fit_lineage_curves()

            # Ensure starts at 0
            for l_idx, lineage in enumerate(self.lineages):
                curve = self.curves[l_idx]
                min_time = np.min(curve.pseudotimes_interp[self.cell_weights[:, l_idx] > 0])
                curve.pseudotimes_interp -= min_time

            # Determine average curves
            shrinkage_percentages, cluster_children, cluster_avg_curves = \
                self.avg_curves()

            # Shrink towards average curves in areas of cells common to all branch lineages
            self.shrink_curves(cluster_children, shrinkage_percentages, cluster_avg_curves)

            self.debug_plot_lineages = False
            self.debug_plot_avg = False

            if self.debug_axes is not None and epoch == num_epochs - 1:  # plot curves
                self.plotter.clusters(self.debug_axes[1, 1], s=2, alpha=0.5)
                self.plotter.curves(self.debug_axes[1, 1], self.curves)

    def construct_initial_curves(self):
        """Constructs lineage principal curves using piecewise linear initialisation"""
        from pcurvepy2 import PrincipalCurve
        piecewise_linear = list()
        distances = list()

        for l_idx, lineage in enumerate(self.lineages):
            # Calculate piecewise linear path
            p = np.stack(self.cluster_centres[lineage.clusters])
            s = np.zeros(p.shape[0])  # TODO

            cell_mask = np.logical_or.reduce(
                np.array([self.cluster_label_indices == k for k in lineage]))
            cells_involved = self.data[cell_mask]

            curve = PrincipalCurve(k=3)
            curve.project_to_curve(cells_involved, points=p)
            d_sq, dist = curve.project_to_curve(self.data, points=curve.points_interp[curve.order])
            distances.append(d_sq)

            # piecewise_linear.append(PrincipalCurve.from_params(s, p))
            piecewise_linear.append(curve)

        self.curves = piecewise_linear
        self.distances = distances

    def get_lineages(self):
        tree = self.construct_mst(self.start_node)

        # Determine lineages by parsing the MST
        branch_clusters = deque()
        def recurse_branches(path, v):
            num_children = len(tree[v])
            if num_children == 0:  # at leaf, add a None token
                return path + [v, None]
            elif num_children == 1:
                return recurse_branches(path + [v], tree[v][0])
            else:  # at branch
                branch_clusters.append(v)
                return [recurse_branches(path + [v], tree[v][i]) for i in range(num_children)]

        def flatten(li):
            if li[-1] is None:  # special None token indicates a leaf
                yield Lineage(li[:-1])
            else:  # otherwise yield from children
                for l in li:
                    yield from flatten(l)

        lineages = recurse_branches([], self.start_node)
        lineages = list(flatten(lineages))
        self.lineages = lineages
        self.branch_clusters = branch_clusters

        self.cluster_lineages = {k: list() for k in range(self.num_clusters)}
        for l_idx, lineage in enumerate(self.lineages):
            for k in lineage:
                self.cluster_lineages[k].append(l_idx)

        if self.debug_level > 0:
            print('Lineages:', lineages)

    def fit_lineage_curves(self):
        """Updates curve using a cubic spline and projection of data"""
        assert self.lineages is not None
        assert self.curves is not None
        distances = list()

        # Calculate principal curves
        for l_idx, lineage in enumerate(self.lineages):
            curve = self.curves[l_idx]

            # Fit principal curve through data
            # Weights are important as they effectively silence points
            # that are not associated with the lineage.
            curve.fit(
                self.data,
                max_iter=1,
                w=self.cell_weights[:, l_idx]
            )

            if self.debug_plot_lineages:
                cell_mask = np.logical_or.reduce(
                    np.array([self.cluster_label_indices == k for k in lineage]))
                cells_involved = self.data[cell_mask]
                self.debug_axes[0, 1].scatter(cells_involved[:, 0], cells_involved[:, 1], s=2, alpha=0.5)
                alphas = curve.pseudotimes_interp
                alphas = (alphas - alphas.min()) / (alphas.max() - alphas.min())
                for i in np.random.permutation(self.data.shape[0])[:50]:
                    path_from = (self.data[i][0], curve.points_interp[i][0])
                    path_to = (self.data[i][1], curve.points_interp[i][1])
                    self.debug_axes[0, 1].plot(path_from, path_to, c='black', alpha=alphas[i])
                self.debug_axes[0, 1].plot(curve.points_interp[curve.order, 0],
                                           curve.points_interp[curve.order, 1], label=str(lineage))

            d_sq, dist = curve.project_to_curve(self.data, curve.points_interp[curve.order])
            distances.append(d_sq)
        self.distances = distances
        if self.debug_plot_lineages:
            self.debug_axes[0, 1].legend()

    def calculate_cell_weights(self):
        """TODO: annotate, this is a translation from R"""
        cell_weights = [self.cluster_labels_onehot[:, self.lineages[l].clusters].sum(axis=1)
                        for l in range(len(self.lineages))]
        cell_weights = np.stack(cell_weights, axis=1)

        d_sq = np.stack(self.distances, axis=1)
        d_ord = np.argsort(d_sq, axis=None)
        w_prob = cell_weights/cell_weights.sum(axis=1, keepdims=True)  # shape (cells, lineages)
        w_rnk_d = np.cumsum(w_prob.reshape(-1)[d_ord]) / w_prob.sum()

        z = d_sq
        z_shape = z.shape
        z = z.reshape(-1)
        z[d_ord] = w_rnk_d
        z = z.reshape(z_shape)
        z_prime = 1 - z ** 2
        z_prime[cell_weights == 0] = np.nan
        w0 = cell_weights.copy()
        cell_weights = z_prime / np.nanmax(z_prime, axis=1, keepdims=True) #rowMins(D) / D
        np.nan_to_num(cell_weights, nan=1, copy=False) # handle 0/0
        # cell_weights[is.na(cell_weights)] <- 0
        cell_weights[cell_weights > 1] = 1
        cell_weights[cell_weights < 0] = 0
        cell_weights[w0 == 0] = 0

        reassign = True
        if reassign:
            # add if z < .5
            cell_weights[z < .5] = 1 #(rowMins(D) / D)[idx]

            # drop if z > .9 and cell_weights < .1
            ridx = (z.max(axis=1) > .9) & (cell_weights.min(axis=1) < .1)
            w0 = cell_weights[ridx]
            z0 = z[ridx]
            w0[(z0 > .9) & (w0 < .1)] = 0 # !is.na(Z0) & Z0 > .9 & W0 < .1
            cell_weights[ridx] = w0

        self.cell_weights = cell_weights

    def avg_curves(self):
        """
        Starting at leaves, calculate average curves for each branch

        :return: shrinkage_percentages, cluster_children, cluster_avg_curves
        """
        cell_weights = self.cell_weights
        shrinkage_percentages = list()
        cluster_children = dict()  # maps cluster to children
        lineage_avg_curves = dict()
        cluster_avg_curves = dict()
        branch_clusters = self.branch_clusters.copy()
        if self.debug_level > 0:
            print('Reversing from leaf to root')
        if self.debug_plot_avg:
            self.plotter.clusters(self.debug_axes[1, 0], s=4, alpha=0.4)

        while len(branch_clusters) > 0:
            # Starting at leaves, find lineages involved in branch
            k = branch_clusters.pop()
            branch_lineages = self.cluster_lineages[k]
            cluster_children[k] = set()
            for l_idx in branch_lineages:  # loop all lineages through branch
                if l_idx in lineage_avg_curves:  # add avg curve
                    curve = lineage_avg_curves[l_idx]
                else:  # or add leaf curve
                    curve = self.curves[l_idx]
                cluster_children[k].add(curve)

            # Calculate the average curve for this branch
            branch_curves = list(cluster_children[k])
            if self.debug_level > 0:
                print(f'Averaging branch @{k} with lineages:', branch_lineages, branch_curves)

            avg_curve = self.avg_branch_curves(branch_curves)
            cluster_avg_curves[k] = avg_curve
            # avg.curve$w <- rowSums(vapply(pcurves, function(p){ p$w }, rep(0,nrow(X))))

            # Calculate shrinkage weights using areas where cells share lineages
            # note that this also captures cells in average curves, since the
            # lineages which are averaged are present in branch_lineages
            common = cell_weights[:, branch_lineages] > 0
            common_mask = common.mean(axis=1) == 1.
            shrinkage_percent = dict()
            for curve in branch_curves:
                shrinkage_percent[curve] = self.shrinkage_percent(curve, common_mask)
            shrinkage_percentages.append(shrinkage_percent)

            # Add avg_curve to lineage_avg_curve for cluster_children
            for l in branch_lineages:
                lineage_avg_curves[l] = avg_curve
            # # check for degenerate case (if one curve won't be
            # # shrunk, then the other curve shouldn't be,
            # # either)
            # new.avg.order <- avg.order
            # all.zero <- vapply(pct.shrink[[i]], function(pij){
            #     return(all(pij == 0))
            # }, TRUE)
            # if(any(all.zero)){
            #     if(allow.breaks){
            #         new.avg.order[[i]] <- NULL
            #         message('Curves for ', ns[1], ' and ',
            #             ns[2], ' appear to be going in opposite ',
            #             'directions. No longer forcing them to ',
            #             'share an initial point. To manually ',
            #             'override this, set allow.breaks = ',
            #             'FALSE.')
            #     }
            #     pct.shrink[[i]] <- lapply(pct.shrink[[i]],
            #         function(pij){
            #             pij[] <- 0
            #             return(pij)
            #         })
            # }
        if self.debug_plot_avg:
            self.debug_axes[1, 0].legend()
        return shrinkage_percentages, cluster_children, cluster_avg_curves

    def shrink_curves(self, cluster_children, shrinkage_percentages, cluster_avg_curves):
        """
        Starting at root, shrink curves for each branch

        Parameters:
            cluster_children:
            shrinkage_percentages:
            cluster_avg_curves:
        :return:
        """
        branch_clusters = self.branch_clusters.copy()
        while len(branch_clusters) > 0:
            # Starting at root, find lineages involves in branch
            k = branch_clusters.popleft()
            shrinkage_percent = shrinkage_percentages.pop()
            branch_curves = list(cluster_children[k])
            cluster_avg_curve = cluster_avg_curves[k]
            if self.debug_level > 0:
                print(f'Shrinking branch @{k} with curves:', branch_curves)

            # Specify the avg curve for this branch
            self.shrink_branch_curves(branch_curves, cluster_avg_curve, shrinkage_percent)

    def shrink_branch_curves(self, branch_curves, avg_curve, shrinkage_percent):
        """
        Shrinks curves through a branch to the average curve.
        
        Arguments:
            branch_curves: list of `PrincipalCurve`s associated with the branch.
            avg_curve: `PrincipalCurve` for average curve.
            shrinkage_percent: percentage shrinkage, in same order as curve.pseudotimes
       
        """
        num_dims_reduced = branch_curves[0].points_interp.shape[1]

        # Go through "child" lineages, shrinking the curves toward the above average
        for curve in branch_curves:  # curve might be an average curve or a leaf curve
            pct = shrinkage_percent[curve]

            s_interp, p_interp, order = curve.unpack_params()
            avg_s_interp, avg_p_interp, avg_order = avg_curve.unpack_params()
            shrunk_curve = np.zeros_like(p_interp)
            for j in range(num_dims_reduced):
                orig = p_interp[order, j]
                avg = np.interp(#interp1d(
                    s_interp[order],
                    avg_s_interp[avg_order],     # x
                    avg_p_interp[avg_order, j])#,  # y
                    # assume_sorted=True,
                    # bounds_error=False,
                    # fill_value='extrapolate',
                    # extrapolate_extrema=True)
                # avg = lin_interpolator#(s_interp[order])
                shrunk_curve[:, j] = (avg * pct + orig * (1 - pct))
            # w <- pcurve$w
            # pcurve = project_to_curve(X, as.matrix(s[pcurve$ord, ,drop = FALSE]), stretch = stretch)
            # pcurve$w <- w
            # self.debug_axes[1, 1].plot(
            #     shrunk_curve[:, 0],
            #     shrunk_curve[:, 1],
            #     label='shrunk', alpha=0.2, c='black')
            curve.project_to_curve(self.data, points=shrunk_curve)
            #     for(jj in seq_along(ns)){
            #         n <- ns[jj]
            #         if(grepl('Lineage',n)){
            #             l.ind <- as.numeric(gsub('Lineage','',n))
            #             pcurves[[l.ind]] <- shrunk[[jj]]
            #         }
            #         if(grepl('average',n)){
            #             a.ind <- as.numeric(gsub('average','',n))
            #             avg.lines[[a.ind]] <- shrunk[[jj]]
            #         }
            #     }
            # }
            # avg.order <- new.avg.order

    def shrinkage_percent(self, curve, common_ind):
        """Determines how much to shrink a curve"""
        # pst <- crv$lambda
        # pts2wt <- pst
        s_interp, order = curve.pseudotimes_interp, curve.order
        # Cosine kernel quartiles:
        x = self.kernel_x
        y = self.kernel_y
        y = (y.sum() - np.cumsum(y)) / sum(y)
        q1 = np.percentile(s_interp[common_ind], 25)
        q3 = np.percentile(s_interp[common_ind], 75)
        a = q1 - 1.5 * (q3 - q1)
        b = q3 + 1.5 * (q3 - q1)
        x = scale_to_range(x, a=a, b=b)
        if q1 == q3:
            pct_l = np.zeros(s_interp.shape[0])
        else:
            pct_l = np.interp(
                s_interp[order],
                x, y
            )

        return pct_l

    def avg_branch_curves(self, branch_curves):
        """branch_lineages is a list of lineages passing through branch"""
        from pcurvepy2 import PrincipalCurve
        # s_interps, p_interps, orders
        num_cells = branch_curves[0].points_interp.shape[0]
        num_dims_reduced = branch_curves[0].points_interp.shape[1]

        # 1. Interpolate all the lineages over the shared time domain
        branch_s_interps = np.stack([c.pseudotimes_interp for c in branch_curves], axis=1)
        max_shared_pseudotime = branch_s_interps.max(axis=0).min()  # take minimum of maximum pseudotimes for each lineage
        combined_pseudotime = np.linspace(0, max_shared_pseudotime, num_cells)
        curves_dense = list()
        for curve in branch_curves:
            lineage_curve = np.zeros((combined_pseudotime.shape[0], num_dims_reduced))
            order = curve.order
            # Linearly interpolate each dimension as a function of pseudotime
            for j in range(num_dims_reduced):
                lin_interpolator = interp1d(
                    curve.pseudotimes_interp[order], # x
                    curve.points_interp[order, j],   # y
                    assume_sorted=True
                )
                lineage_curve[:, j] = lin_interpolator(combined_pseudotime)
            curves_dense.append(lineage_curve)

        curves_dense = np.stack(curves_dense, axis=1)  # (n, L_b, J)

        # 2. Average over these curves and project the data onto the result
        avg = curves_dense.mean(axis=1)  # avg is already "sorted"
        avg_curve = PrincipalCurve()
        avg_curve.project_to_curve(self.data, points=avg)
        # avg_curve.pseudotimes_interp -= avg_curve.pseudotimes_interp.min()
        if self.debug_plot_avg:
            self.debug_axes[1, 0].plot(avg[:, 0], avg[:, 1], c='blue', linestyle='--', label='average', alpha=0.7)
            _, p_interp, order = avg_curve.unpack_params()
            self.debug_axes[1, 0].plot(p_interp[order, 0], p_interp[order, 1], c='red', label='data projected', alpha=0.7)

        # avg.curve$w <- rowSums(vapply(pcurves, function(p){ p$w }, rep(0,nrow(X))))
        return avg_curve

    @property
    def unified_pseudotime(self):
        pseudotime = np.zeros_like(self.curves[0].pseudotimes_interp)
        for l_idx, lineage in enumerate(self.lineages):
            curve = self.curves[l_idx]
            cell_mask = np.logical_or.reduce(
                np.array([self.cluster_label_indices == k for k in lineage]))
            pseudotime[cell_mask] = curve.pseudotimes_interp[cell_mask]
        return pseudotime

    def list_lineages(self, cluster_to_label):
        for lineage in self.lineages:
            print(', '.join([
                cluster_to_label[l] for l in lineage
            ]))


r"""
External algorithms and specialized methods collection.

This module provides access to a comprehensive collection of external algorithms
and specialized methods that extend omicverse's core functionality. These include
state-of-the-art methods for spatial transcriptomics, trajectory inference,
multi-omics integration, and advanced single-cell analysis.

Algorithm categories:
    Spatial transcriptomics:
        - STAGATE_pyG: Spatial domain identification
        - STAligner: Multi-sample spatial alignment
        - GraphST: Graph-based spatial analysis
        - PROST: Pattern recognition for spatial data
        - spatrio: Spatial transcriptomics analysis
        - CAST: Cellular spatial organization
        - starfysh: Spatial factorization
        
    Trajectory inference:
        - STT: Spatial transition tensor
        - VIA: Velocity integration and annotation
        - cytotrace2: Developmental potential scoring
        
    Multi-omics integration:
        - mofapy2: Multi-Omics Factor Analysis
        - GNTD: Graph-regularized non-negative tensor decomposition
        - scMulan: Multi-modal analysis
        
    Single-cell methods:
        - tosica: Trajectory inference and cell fate
        - cnmf: Consensus non-negative matrix factorization
        - CEFCON: Cell fate controller networks
        - cellanova: Cell type-aware analysis
        - BINARY: Binary classification methods
        - flowsig: Flow cytometry-style analysis
        - scdiffusion: Diffusion-based methods
        
    Network analysis:
        - PyWGCNA: Weighted gene co-expression networks
        - commot: Cell-cell communication
        
    Data integration:
        - scanorama: Batch correction and integration
        - scSLAT: Single-cell alignment
        
    Spatial methods:
        - spaceflow: Spatial flow analysis
        - gaston: Spatial depth estimation

Features:
    - Optional dependencies: Algorithms are loaded as needed
    - Consistent interfaces: Wrapped for omicverse compatibility
    - Performance optimized: GPU acceleration where available
    - Well documented: Each method includes comprehensive examples

Notes:
    - Some algorithms require additional dependencies
    - GPU methods require appropriate hardware and drivers
    - All methods are thoroughly tested and validated
    - Regular updates incorporate latest algorithmic advances
"""
from . import (scSLAT,CEFCON,mofapy2,GNTD,spaceflow,STT,
               tosica,STAGATE_pyG,STAligner,spatrio,PROST,cytotrace2,
               GraphST,commot,cnmf,starfysh,scMulan,flowsig,PyWGCNA,
               CAST,scanorama,scdiffusion,BINARY,cellanova,VIA,gaston,pyscenic
               )
__all__ = [
    'scSLAT',
    'CEFCON',
    'mofapy2',
    'GNTD',
    'spaceflow',
    'STT',
    'tosica',
    'STAGATE_pyG',
    'STAligner',
    'spatrio',
    'PROST',
    'cytotrace2',
    'GraphST',
    'commot',
    'cnmf',
    'starfysh',
    'scMulan',
    'flowsig',
    'PyWGCNA',
    'CAST',
    'scanorama',
    'scdiffusion',
    'BINARY',
    'cellanova',
    'VIA',
    'gaston',
    'pyscenic',
]




import numpy
import time
import sys
import sklearn.neighbors
import scipy.optimize

from typing import Optional

def icp(a, b,
        max_time:Optional[int] = 4
    ):
    """
    Iterative Closest Point (ICP) algorithm
    
    Parameters
    ----------
    a
        dataset a, whose shape is [n,2]
    b
        dataset a, whose shape is [m,2]
    max_time
        max iter times
    
    Return
    ----------
    T_opt
        optimal transform matrix, please see https://docs.opencv.org/3.4/da/d6e/tutorial_py_geometric_transformations.html
    error_max
        error
    
    Refer
    ----------
    https://stackoverflow.com/questions/20120384/iterative-closest-point-icp-implementation-on-python
    """
    import cv2
    def res(p,src,dst):
        T = numpy.matrix([[numpy.cos(p[2]),-numpy.sin(p[2]),p[0]],
        [numpy.sin(p[2]), numpy.cos(p[2]),p[1]],
        [0 ,0 ,1 ]])
        n = numpy.size(src,0)
        xt = numpy.ones([n,3])
        xt[:,:-1] = src
        xt = (xt*T.T).A
        d = numpy.zeros(numpy.shape(src))
        d[:,0] = xt[:,0]-dst[:,0]
        d[:,1] = xt[:,1]-dst[:,1]
        r = numpy.sum(numpy.square(d[:,0])+numpy.square(d[:,1]))
        return r

    def jac(p,src,dst):
        T = numpy.matrix([[numpy.cos(p[2]),-numpy.sin(p[2]),p[0]],
        [numpy.sin(p[2]), numpy.cos(p[2]),p[1]],
        [0 ,0 ,1 ]])
        n = numpy.size(src,0)
        xt = numpy.ones([n,3])
        xt[:,:-1] = src
        xt = (xt*T.T).A
        d = numpy.zeros(numpy.shape(src))
        d[:,0] = xt[:,0]-dst[:,0]
        d[:,1] = xt[:,1]-dst[:,1]
        dUdth_R = numpy.matrix([[-numpy.sin(p[2]),-numpy.cos(p[2])],
                            [ numpy.cos(p[2]),-numpy.sin(p[2])]])
        dUdth = (src*dUdth_R.T).A
        g = numpy.array([  numpy.sum(2*d[:,0]),
                        numpy.sum(2*d[:,1]),
                        numpy.sum(2*(d[:,0]*dUdth[:,0]+d[:,1]*dUdth[:,1])) ])
        return g
    
    def hess(p,src,dst):
        #import cv2
        n = numpy.size(src,0)
        T = numpy.matrix([[numpy.cos(p[2]),-numpy.sin(p[2]),p[0]],
        [numpy.sin(p[2]), numpy.cos(p[2]),p[1]],
        [0 ,0 ,1 ]])
        n = numpy.size(src,0)
        xt = numpy.ones([n,3])
        xt[:,:-1] = src
        xt = (xt*T.T).A
        d = numpy.zeros(numpy.shape(src))
        d[:,0] = xt[:,0]-dst[:,0]
        d[:,1] = xt[:,1]-dst[:,1]
        dUdth_R = numpy.matrix([[-numpy.sin(p[2]),-numpy.cos(p[2])],[numpy.cos(p[2]),-numpy.sin(p[2])]])
        dUdth = (src*dUdth_R.T).A
        H = numpy.zeros([3,3])
        H[0,0] = n*2
        H[0,2] = numpy.sum(2*dUdth[:,0])
        H[1,1] = n*2
        H[1,2] = numpy.sum(2*dUdth[:,1])
        H[2,0] = H[0,2]
        H[2,1] = H[1,2]
        d2Ud2th_R = numpy.matrix([[-numpy.cos(p[2]), numpy.sin(p[2])],[-numpy.sin(p[2]),-numpy.cos(p[2])]])
        d2Ud2th = (src*d2Ud2th_R.T).A
        H[2,2] = numpy.sum(2*(numpy.square(dUdth[:,0])+numpy.square(dUdth[:,1]) + d[:,0]*d2Ud2th[:,0]+d[:,0]*d2Ud2th[:,0]))
        return H
    
    t0 = time.time()
    init_pose = (0,0,0)
    src = numpy.array([a.T], copy=True).astype(numpy.float32)
    dst = numpy.array([b.T], copy=True).astype(numpy.float32)
    Tr = numpy.array([[numpy.cos(init_pose[2]),-numpy.sin(init_pose[2]),init_pose[0]],
                   [numpy.sin(init_pose[2]), numpy.cos(init_pose[2]),init_pose[1]],
                   [0,                    0,                   1          ]])
    print("src",numpy.shape(src))
    print("Tr[0:2]",numpy.shape(Tr[0:2]))
    src = cv2.transform(src, Tr[0:2])
    p_opt = numpy.array(init_pose)
    T_opt = numpy.array([])
    error_max = sys.maxsize
    first = False
    while not(first and time.time() - t0 > max_time):
        distances, indices = sklearn.neighbors.NearestNeighbors(n_neighbors=1, algorithm='auto',p = 3).fit(dst[0]).kneighbors(src[0])
        p = scipy.optimize.minimize(res,[0,0,0],args=(src[0],dst[0, indices.T][0]),method='Newton-CG',jac=jac,hess=hess).x
        T  = numpy.array([[numpy.cos(p[2]),-numpy.sin(p[2]),p[0]],[numpy.sin(p[2]), numpy.cos(p[2]),p[1]]])
        p_opt[:2]  = (p_opt[:2]*numpy.matrix(T[:2,:2]).T).A       
        p_opt[0] += p[0]
        p_opt[1] += p[1]
        p_opt[2] += p[2]
        src = cv2.transform(src, T)
        Tr = (numpy.matrix(numpy.vstack((T,[0,0,1])))*numpy.matrix(Tr)).A
        error = res([0,0,0],src[0],dst[0, indices.T][0])

        if error < error_max:
            error_max = error
            first = True
            T_opt = Tr

    p_opt[2] = p_opt[2] % (2*numpy.pi)
    
    return T_opt, error_max


from typing import List, Tuple, Optional
import numpy as np
from anndata import AnnData
import scanpy as sc
import pandas as pd
from tqdm import tqdm
#
from scipy.optimize import nnls
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import normalize
from scipy.stats import pearsonr
#from .helper import intersect, kl_divergence_backend, to_dense_array, extract_exp, extract_data_matrix, process_anndata, construct_graph, distances_cal, dist_cal, scale_num
from .helper import *
    
def ot_alignment(
    adata1: AnnData,  
    adata2: AnnData,
    alpha: float = 0.1, 
    dissimilarity: str = 'scaled_euc',
    n_components: int = 30,  
    G_init = None,
    p_distribution = None, 
    q_distribution = None, 
    numItermax: int = 200, 
    norm: str = 'l2', 
    backend = None,  
    return_obj: bool = False,
    verbose: bool = True, 
    k: int = 10,
    graph_mode: str = "connectivity",
    aware_spatial: bool = True,
    aware_multi: bool = True,
    aware_power: int = 2,
    **kwargs) :
    
    """
    Calculates and returns optimal alignment spot and single cell data. 
    
    Args:
        adata1: Spatial transcriptomic data.
        adata2: Single cell multi-omics data.
        alpha:  Alignment tuning parameter. Note: 0 <= alpha <= 1.
        dissimilarity: Expression dissimilarity measure: "count_euc", "data_euc", "scaled_euc", "pca_euc", "kl".
        n_components: Number of pca dimensions selected when processing input data
        G_init (array-like, optional): Initial mapping to be used in FGW-OT, otherwise default is uniform mapping.
        p_distribution (array-like, optional): Distribution of adata1 spots, otherwise default is uniform.
        q_distribution (array-like, optional): Distribution of adata2 cells, otherwise default is uniform.
        numItermax: Max number of iterations.
        norm: Determines what sort of normalization to run on low dimensional representation of adata2, Default="l2".
        backend: Type of backend to run calculations.
        return_obj: Determines whether to use returns objective function output of FGW-OT.
        verbose: Prints loss when optimizing the optimal transport formulation. Default=True.
        k: Number of neighbors to be used when constructing kNN graph.
        graph_mode: "connectivity" or "distance". Determines whether to use a connectivity graph or a distance graph.Default="connectivity".
        aware_power: Type aware parameter. The greater the parameter, the greater the distance between different areas/types of spots/cells in the graph.
        aware_spatial: Determines whether to adjust the distance between spots according to areas (or other meta info)
        aware_multi: Determines whether to adjust the distance between cells according to types (or other meta info)
        
   
    Returns:
        - Alignment results between spots and cells.

        If ``return_obj = True``, additionally returns:
        
        - Objective function output of FGW-OT.
    """
    import ot
    if backend is None:
        backend=ot.backend.NumpyBackend()
    nx = backend
    n1 = adata1.shape[0]
    n2 = adata2.shape[0]
    
    # Calculate expression dissimilarity
    if dissimilarity == 'pca_euc':
        print('Calculating dissimilarity using euclidean distance on pca low dimensional...')
        # Merge tow anndata
        ad1=adata1.copy()
        ad1.X=ad1.layers['counts']
        ad2=adata2.copy()
        ad2.X=ad2.layers['counts']
        adata_merge = ad1.concatenate(ad2,index_unique=None)
        # Process the merged anndata
        adata_merge = process_anndata(adata_merge,ndims = n_components)
        reduc_spot = adata_merge.obsm["X_pca"][0:n1,]
        reduc_single = adata_merge.obsm["X_pca"][n1:n1+n2,]
        # reduc_spot = adata_merge.X[0:n1,]
        # reduc_single = adata_merge.X[n1:n1+n2,]
        M = ot.dist(reduc_spot, reduc_single, metric='euclidean')
        M /= M.max()
        M = nx.from_numpy(M)
        del ad1,ad2

    if dissimilarity == 'scaled_euc':
        print('Calculating dissimilarity using euclidean distance on scaled data...')
        ad1=adata1.copy()
        ad1.X=ad1.layers['counts']
        ad2=adata2.copy()
        ad2.X=ad2.layers['counts']
        adata_merge = ad1.concatenate(ad2,index_unique=None)
        # Process the merged anndata
        adata_merge = process_anndata(adata_merge,ndims = n_components,scale=True,pca=False)
        reduc_spot = adata_merge.X[0:n1,]
        reduc_single = adata_merge.X[n1:n1+n2,]
        M = ot.dist(reduc_spot, reduc_single, metric='euclidean')
        M /= M.max()
        M = nx.from_numpy(M)
        del ad1,ad2

    if dissimilarity == 'data_euc':
        print('Calculating dissimilarity using euclidean distance on normalized data...')
        # Merge tow anndata
        ad1=adata1.copy()
        ad1.X=ad1.layers['counts']
        ad2=adata2.copy()
        ad2.X=ad2.layers['counts']
        adata_merge = ad1.concatenate(ad2,index_unique=None)
        # Process the merged anndata
        adata_merge = process_anndata(adata_merge,ndims = n_components,scale=False,pca=False)
        reduc_spot = adata_merge.X[0:n1,]
        reduc_single = adata_merge.X[n1:n1+n2,]
        M = ot.dist(reduc_spot, reduc_single, metric='euclidean')
        M /= M.max()
        M = nx.from_numpy(M)
        del ad1,ad2

    if dissimilarity == 'count_euc':
        print('Calculating dissimilarity using euclidean distance on count data...')
        # Calculate expression dissimilarity
        ad1=adata1.copy()
        ad1.X=ad1.layers['counts']
        ad2=adata2.copy()
        ad2.X=ad2.layers['counts']
        #adata_merge = ad1.concatenate(ad2,index_unique=None)
        A_X, B_X = nx.from_numpy(to_dense_array(extract_data_matrix(ad1,rep = None))), nx.from_numpy(to_dense_array(extract_data_matrix(ad2,rep = None)))
        M = ot.dist(A_X,B_X)
        M /= M.max()
        M = nx.from_numpy(M)
        del ad1,ad2
        
    if dissimilarity == 'kl':
        print('Computing dissimilarity using kl divergence...')
        ad1=adata1.copy()
        ad1.X=ad1.layers['counts']
        ad2=adata2.copy()
        ad2.X=ad2.layers['counts']
        A_X, B_X = nx.from_numpy(to_dense_array(extract_data_matrix(ad1,rep = None))), nx.from_numpy(to_dense_array(extract_data_matrix(ad2,rep = None)))
        s_A = A_X + 0.01
        s_B = B_X + 0.01
        M = kl_divergence_backend(s_A, s_B)
        M = nx.from_numpy(M) #r:ncellA*c:ncellB
        del ad1,ad2

    # Construct the graph
    location_array = adata1.obsm['spatial'].values
    reduction_array = adata2.obsm['reduction'].values
    reduction_array =normalize(reduction_array, norm=norm, axis=1)
    #Xgraph = construct_graph(location_array, k=k, mode=graph_mode, type_aware=spot_meta, aware_power=0)###break up
    print('Constructing '+str(graph_mode)+"...")
    print('k = '+str(k))
    Xgraph = construct_graph(location_array, k=k, mode=graph_mode)
    ygraph = construct_graph(reduction_array, k=k, mode=graph_mode)

    # Adjust the distance according to meta info
    type_aware1 = None
    type_aware2 = None
    if aware_spatial:
        print('aware_spatial = True')
        type_aware_dict = {
        'spot': pd.Series(adata1.obs.index.tolist(),index=adata1.obs.index.tolist()),
        'spot_type': pd.Series(adata1.obs['type'],index=adata1.obs.index.tolist())
        }
        type_aware1 = pd.DataFrame(type_aware_dict)
    if aware_multi:
        print('aware_multi = True')
        type_aware_dict = {
        'single': pd.Series(adata2.obs.index.tolist(),index=adata2.obs.index.tolist()),
        'single_type': pd.Series(adata2.obs['type'],index=adata2.obs.index.tolist())
        }
        type_aware2 = pd.DataFrame(type_aware_dict)
    print('aware power = '+str(aware_power))
    Cx = distances_cal(Xgraph,type_aware=type_aware1, aware_power=aware_power)
    Cy = distances_cal(ygraph,type_aware=type_aware2, aware_power=aware_power)
    Cx = nx.from_numpy(Cx)
    Cy = nx.from_numpy(Cy)

    # Init distributions
    if p_distribution is None:
        p = np.ones((n1,)) / n1
        p = nx.from_numpy(p)
    else:
        p = nx.from_numpy(p_distribution)
    if q_distribution is None:
        q = np.ones((n2,)) / n2
        q = nx.from_numpy(q)
    else:
        q = nx.from_numpy(q_distribution)

    # Run OT
    if G_init is not None:
        G_init = nx.from_numpy(G_init)
   
    print('Running OT...')
    print('alpha = '+str(alpha))
    pi, logw = my_ot(M, Cx, Cy, p, q, G_init = G_init, loss_fun='square_loss', 
                     alpha= alpha, log=True, numItermax=numItermax,verbose=verbose)
    pi = nx.to_numpy(pi)
    print("OT done!")
    #obj = nx.to_numpy(logw['fgw_dist'])
    out_data = pd.DataFrame(pi)
    out_data.columns = adata2.obsm['reduction'].index
    out_data.index =  adata1.obsm['spatial'].index

    # Filter the results
    out_data['spot']=out_data.index
    out_data = out_data.melt(id_vars=['spot'], var_name ='cell', value_name='value')
    out_data = out_data.sort_values(by="value",ascending=False)

    if return_obj:
        return pi, out_data
    return out_data



def assign_coord(
    adata1: AnnData,
    adata2: AnnData,
    out_data: pd.DataFrame, 
    non_zero_probabilities: bool = True,
    no_repeated_cells: bool = True,
    top_num: int = None,
    expected_num: pd.DataFrame = None,
    random: bool = False,
    normalize: bool = True,
    ) :
    """
    Assign coordinates for single cells according to the output of ot alignment.
    
    Args:
        adata1: Spatial transcriptomic data.
        adata2: Single cell multi-omics data.
        out_data:  Alignment results between spots and cells from previous step.
        non_zero_probabilities: Determines whether to remove 0 frome alignment results. Default=True.
        no_repeated_cells: Determines whether to allow a cell to be used multiple times when allocating coordinates. Default=True.
        top_num: The maximum number of cells allocated in a spot, Default=5.
        expected_num: DataFrame specifying the expected number of cells per spot. Default is None.
        random: Determines whether to randomly assign cell coordinates or assign coordinates based on pearson correlation coefficient. Default=False.
        expected_num: DataFrame specifying the expected number of cells per spot. Default is None.
    
    Returns:
    Returns:
        - Spatial assignment of cells.
    """
    
    print('Assigning spatial coordinates to cells...')
    print('random = '+str(random))
    if top_num is not None:
        print("Maximum number of cells assigned to spots: "+str(top_num))
        cell_num = {spot: top_num for spot in pd.unique(adata1.obs_names)}
    if expected_num is not None:
        print("Determine the number of cells allocated to each spot based on the input information")
        cell_num = expected_num['cell_num'].to_dict()

    if non_zero_probabilities :
        out_data = out_data[out_data['value']>0]
    if no_repeated_cells :   
        out_data = out_data.sort_values(by="value",ascending=False)
        out_data = out_data[out_data.duplicated('cell') == False]

    if normalize:
        print('Normalizing the alignment results...')
        
        adata1_copy=adata1.copy()
        adata1_copy.X=adata1_copy.layers['counts']
        adata2_copy=adata2.copy()
        adata2_copy.X=adata2_copy.layers['counts']

        adata1_copy = process_anndata(adata1_copy,ndims=50,scale=False,pca=False)
        adata2_copy = process_anndata(adata2_copy,ndims=50,scale=False,pca=False)
    else:
        adata1_copy=adata1.copy()
        adata2_copy=adata2.copy()
        
    common_genes = intersect(adata1_copy.var.index, adata2_copy.var.index)
    adata1_copy = adata1_copy[:, common_genes]
    adata2_copy = adata2_copy[:, common_genes]
    
    adata1_copy = adata1_copy[adata1_copy.obs.index.isin(pd.unique(out_data.spot))]
    adata2_copy = adata2_copy[adata2_copy.obs.index.isin(pd.unique(out_data.cell))]

    res2 = pd.DataFrame(columns=adata2_copy.var_names, index=adata2_copy.obs['type'].astype('category').cat.categories)
    
    for clust in adata2_copy.obs['type'].astype('category').cat.categories: 
        res2.loc[clust] = adata2_copy[adata2_copy.obs['type'].isin([clust]),:].X.mean(0)
    res2 = np.array(res2)
    res2 = np.transpose(res2)  
    data = extract_exp(adata1_copy)
    ratio_df = pd.DataFrame(columns=adata2_copy.obs['type'].astype('category').cat.categories, index=pd.unique(out_data.spot))   
    for spot in pd.unique(out_data.spot):
        res1 = data.loc[spot]
        res1 = res1.T.values  # (gene_num, 1)
        res1 = res1.reshape(res1.shape[0],)
        ratio_sub = nnls(res2, res1)[0]
        
        ratio_sum = np.sum([ratio_sub], axis=1)[0]
        if ratio_sum == 0:  
            ratio_sub = [0] * len(ratio_sub)
        else:
            ratio_sub = (ratio_sub / ratio_sum).tolist()
        ratio_sub = np.round(np.array(ratio_sub)*cell_num[spot])
        ratio_df.loc[spot] = ratio_sub

    meta1_dict = {
        'spot': pd.Series(adata1.obs.index.tolist(),index=adata1.obs.index.tolist()),
        'spot_type': pd.Series(adata1.obs['type'],index=adata1.obs.index.tolist())
    }
    meta1 = pd.DataFrame(meta1_dict)
    out_data = pd.merge(out_data, meta1, on='spot',how="left")

    meta2_dict = {
        'spot': pd.Series(adata2.obs.index.tolist(),index=adata2.obs.index.tolist()),
        'spot_type': pd.Series(adata2.obs['type'],index=adata2.obs.index.tolist())
    }
    meta2 = pd.DataFrame(meta2_dict)
    meta2.columns = ['cell','cell_type']
    out_data = pd.merge(out_data, meta2, on='cell',how="left")


    # Assign cell to spot
    decon_df = pd.DataFrame(columns=out_data.columns)
    for spot in pd.unique(out_data.spot):
        spot_ratio = ratio_df.loc[spot]
        spot_ot = out_data.loc[out_data['spot'] == spot]
        decon_spot1 = pd.DataFrame(columns=out_data.columns)
        for cluster_id in range(0,len(spot_ratio)):
            cluster = spot_ratio.index[cluster_id]
            decon_num = spot_ratio[cluster_id]
            decon_spot_ot = spot_ot.loc[spot_ot['cell_type'] == cluster][0:int(decon_num)]
            decon_spot1 = pd.concat([decon_spot1,decon_spot_ot])
        decon_num = decon_spot1.shape[0]
        if decon_num < cell_num[spot]:
            rest_spot_ot = spot_ot.drop(decon_spot1.index)
            rest_spot_ot = rest_spot_ot.sort_values(by="value",ascending=False)
            decon_spot2 = rest_spot_ot.iloc[0:(cell_num[spot]-decon_num)]
            decon_spot = pd.concat([decon_spot1,decon_spot2])
        elif decon_num > 0 :
            decon_spot = decon_spot1
        decon_df = pd.concat([decon_df,decon_spot])
        
    out_data = decon_df.groupby('spot').apply(lambda x: x.nlargest(cell_num[x.name], 'value'))

    # Adjust cell coord
    if random:
        ## Calculate radius
        coord = adata1.obsm['spatial'].copy()
        a = coord[['x', 'y']].to_numpy()
        nbrs = NearestNeighbors(n_neighbors=2, algorithm='ball_tree').fit(a)
        distances, indices = nbrs.kneighbors(coord)
        radius = distances[:, -1] / 2
        radius = radius.tolist()
        coord['spot'] = coord.index
        
        ## Calculate coord randomly  
        if len(list(out_data.index.names)) > 1:
            out_data.index = out_data.index.droplevel()
        df_meta = pd.merge(out_data,coord,on='spot',how="left")
        all_coord = df_meta[['x', 'y']].to_numpy()

        mean_radius = np.mean(radius)
        all_radius = [mean_radius] * all_coord.shape[0]

        length = np.random.uniform(0, all_radius)
        angle = np.pi * np.random.uniform(0, 2, all_coord.shape[0])
        x = all_coord[:, 0] + length * np.cos(angle)
        y = all_coord[:, 1] + length * np.sin(angle)
        cell_coord = {'Cell_xcoord': np.around(x, 2).tolist(), 'Cell_ycoord': np.around(y, 2).tolist()}
        df_cc = pd.DataFrame(cell_coord)
        df_meta = pd.concat([df_meta, df_cc], axis=1)
    else:
        ## Calculate radius
        coord = adata1.obsm['spatial'].copy()
        a = coord[['x', 'y']].to_numpy()
        nbrs = NearestNeighbors(n_neighbors=2, algorithm='ball_tree').fit(a)
        distances, indices = nbrs.kneighbors(coord.values)
        radius = distances[:, -1] / 2
        radius = radius.tolist()
        mean_radius = np.mean(radius)

        if len(list(out_data.index.names)) > 1:
            out_data.index = out_data.index.droplevel()

        ## Calculate dist of two spots
        dist = pd.DataFrame(columns=coord.index,index = coord.index)
        dist['spot1']=dist.index
        dist = dist.melt(id_vars=['spot1'], var_name ='spot2', value_name='value')
        coord1 = coord.copy()
        coord1.columns = ['x1','y1']
        coord1['spot1']=coord1.index
        coord2 = coord.copy()
        coord2.columns = ['x2','y2']
        coord2['spot2']=coord2.index
        dist = pd.merge(dist, coord1, on='spot1',how="left")
        dist = pd.merge(dist, coord2, on='spot2',how="left")
        
        ## Filter dist dataframe with n_neighber to speed up
        a = coord[['x', 'y']].to_numpy()
        nbrs = NearestNeighbors(n_neighbors=20, algorithm='ball_tree').fit(a)
        distances, indices = nbrs.kneighbors(coord.values)
        indices = pd.DataFrame(indices)
        nn_dic = {indices.index.tolist()[i]:coord.index.tolist()[i] for i in range(coord.shape[0])}
        indices = indices.replace(nn_dic)
        indices = indices.rename(columns={0:'spot1'})
        nn_dist = indices.melt(id_vars=['spot1'], var_name ='num', value_name='spot2')
        nn_dist["spot_pair"] = nn_dist["spot1"]+nn_dist["spot2"]
        dist["spot_pair"] = dist["spot1"]+dist["spot2"]
        dist = dist[dist['spot_pair']. isin(nn_dist["spot_pair"])]
        dist = dist.drop(['spot_pair'],axis=1)
        
        ## Filter spot that asigned no cells
        spot_num = coord.shape[0]
        spot_has_cell_num = len(np.unique(out_data.spot))
        dist = dist[dist.spot1.isin(pd.unique(out_data.spot))&dist.spot2.isin(pd.unique(out_data.spot))]
        print('There are '+str(spot_num)+' spots and '+str(spot_has_cell_num)+' of them were assigned cells.')
        
        
        dist_dis = []
        for i in dist.index:
            dist_dis.append(dist_cal(dist.loc[i].x1,dist.loc[i].y1,dist.loc[i].x2,dist.loc[i].y2))
        dist.value = dist_dis

        ## Select nearest neighbors of each spot
        dist_closest = dist[dist['value'] > 0]
        dist_closest = dist_closest[dist_closest['value'] < 1.5*min(dist_closest.value)]
        num_closest = pd.value_counts(dist_closest.spot2).max()

        ## Make gene expression data of mappeds spot and cells
        exp_adata2 = extract_exp(adata2_copy)
        exp_adata2_copy = exp_adata2.copy()
        exp_adata2_copy['cell'] = exp_adata2_copy.index
        exp_mapped = pd.merge(out_data[['spot','cell']], exp_adata2_copy, on='cell',how="left")
        exp_mapped = exp_mapped.drop(['cell'],axis=1)
        exp_mapped = exp_mapped.groupby('spot')[np.setdiff1d(exp_mapped.columns, 'spot')].mean()

        df_meta = pd.DataFrame(columns = list(out_data.columns)+['x','y','Cell_xcoord','Cell_ycoord'] )
        for each_spot in tqdm(np.unique(out_data.spot)):
            each_spot_x = coord.loc[each_spot].x
            each_spot_y = coord.loc[each_spot].y

            ### Claculate dist to neighbors of each spot
            dist_of_each_spot = dist[dist.spot1==each_spot]
            dist_of_each_spot = dist_of_each_spot[dist_of_each_spot['value'] < 3*mean_radius]
            dist_of_each_spot = dist_of_each_spot[dist_of_each_spot['value'] > 0]
            if dist_of_each_spot.shape[0] == 0 :
                dist_of_each_spot = dist[dist.spot1==each_spot]
                
            ### Add pseudo cell when neighbors are insufficient
            if dist_of_each_spot.shape[0] < num_closest:
                x_sum = (dist_of_each_spot.shape[0]+1)*each_spot_x
                y_sum = (dist_of_each_spot.shape[0]+1)*each_spot_y
                x_pseudo = x_sum-sum(dist_of_each_spot.x2)
                y_pseudo = y_sum-sum(dist_of_each_spot.y2)
                value_pseudo = dist_cal(each_spot_x,each_spot_y,x_pseudo,y_pseudo)
                pseudo_data = [each_spot,each_spot,value_pseudo,each_spot_x,each_spot_y,x_pseudo,y_pseudo]
                pseudo_data = pd.DataFrame(pseudo_data,columns=[each_spot],index=dist_of_each_spot.columns).T
                dist_of_each_spot =  pd.concat([dist_of_each_spot,pseudo_data])
                
            if dist_of_each_spot.shape[0] > num_closest:
                dist_of_each_spot.nsmallest(num_closest,"value",keep='all')
            

            ### Extract ot output of each spot
            spot_cell_ot = out_data[out_data.spot==each_spot].copy()
            exp_mapped = exp_mapped[exp_adata2.columns]
            spot_cell_ot.loc[:,'Cell_xcoord'] = each_spot_x
            spot_cell_ot.loc[:,'Cell_ycoord'] = each_spot_y
            spot_cell_ot.loc[:,'x'] = each_spot_x
            spot_cell_ot.loc[:,'y'] = each_spot_y

            ### Align cells according to pearson correlation coefficient calculated with neighbor spots
            
            for cell_self in spot_cell_ot.cell:
                exp_cell = exp_adata2.loc[cell_self].values
                neighbor_pearson = []
                for neighbor_spot in dist_of_each_spot.spot2:
                    exp_spot = exp_mapped.loc[neighbor_spot].values
                    pc = pearsonr(exp_cell,exp_spot)
                    neighbor_pearson.append(pc[0])

                if len(neighbor_pearson)>2:
                    neighbor_pearson_scaled = scale_num(neighbor_pearson)###scale to 0-1
                elif len(neighbor_pearson)>1:
                    neighbor_pearson_scaled = neighbor_pearson
                else:
                    neighbor_pearson_scaled = 0
                    
                dist_of_each_spot=dist_of_each_spot.copy()
                
                dist_of_each_spot.loc[:,'x_difference'] = dist_of_each_spot.x2 - dist_of_each_spot.x1
                dist_of_each_spot.loc[:,'y_difference'] = dist_of_each_spot.y2 - dist_of_each_spot.y1
                
                x_map = np.mean(dist_of_each_spot.x_difference * neighbor_pearson_scaled + dist_of_each_spot.x1)
                y_map = np.mean(dist_of_each_spot.y_difference * neighbor_pearson_scaled + dist_of_each_spot.y1)
                spot_cell_ot.loc[spot_cell_ot.cell==cell_self,'Cell_xcoord'] = x_map
                spot_cell_ot.loc[spot_cell_ot.cell==cell_self,'Cell_ycoord'] = y_map

            ### Adjust coord to make cells more distributed
            if spot_cell_ot.shape[0] > 1:
                x_midpoint = np.mean(spot_cell_ot.Cell_xcoord)
                y_midpoint = np.mean(spot_cell_ot.Cell_ycoord)
                spot_cell_ot.Cell_xcoord = spot_cell_ot.Cell_xcoord + each_spot_x - x_midpoint
                spot_cell_ot.Cell_ycoord = spot_cell_ot.Cell_ycoord + each_spot_y - y_midpoint
                x_dif = spot_cell_ot.Cell_xcoord - each_spot_x
                y_dif = spot_cell_ot.Cell_ycoord - each_spot_y
                #### Restrict coord to the scope of the spot
                squ = x_dif * x_dif + y_dif * y_dif
                # if spot_cell_ot.shape[0] > 2:
                ratio = mean_radius/max(squ ** 0.5)
                spot_cell_ot.Cell_xcoord = x_dif * ratio + each_spot_x
                spot_cell_ot.Cell_ycoord = y_dif * ratio + each_spot_y
                # else :
                #     ratio = mean_radius/max(squ ** 0.5)
                #     spot_cell_ot.Cell_xcoord = x_dif * ratio /2 + each_spot_x
                #     spot_cell_ot.Cell_ycoord = y_dif * ratio /2 + each_spot_y
                
            df_meta = pd.concat([df_meta, spot_cell_ot])
            
    print('Assignment done!')
            
    return df_meta

def my_ot(M, C1, C2, p, q, G_init = None, loss_fun='square_loss', alpha=0.1, 
          armijo=False, log=False,numItermax=200,numItermaxEmd=10e6, use_gpu = False, **kwargs):
    """
    Adapted fused_gromov_wasserstein with G_init (inital mapping).
    Also added capability of utilizing different POT backends to speed up computation.
    
    For more info, see: https://pythonot.github.io/gen_modules/ot.gromov.html
    """
    import ot
    p, q = ot.utils.list_to_array(p, q)
    p0, q0, C10, C20, M0 = p, q, C1, C2, M
    nx = ot.backend.get_backend(p0, q0, C10, C20, M0)
    constC, hC1, hC2 = ot.gromov.init_matrix(C1, C2, p, q, loss_fun)

    if G_init is None:
        G0 = p[:, None] * q[None, :]
    else:
        G0 = (1/nx.sum(G_init)) * G_init
        if use_gpu:
            G0 = G0.cuda()
    def f(G):
        return ot.gromov.gwloss(constC, hC1, hC2, G)
    def df(G):
        return ot.gromov.gwggrad(constC, hC1, hC2, G)
    if log:
        if ot.__version__ < '0.9.0':
            res, log = ot.gromov.cg(p, q, (1 - alpha) * M, alpha, f, df, G0, armijo=armijo, C1=C1, C2=C2, constC=constC, log=True, numItermaxEmd=numItermaxEmd, **kwargs)
        else:
            res, log = ot.optim.cg(p, q, (1 - alpha) * M, alpha, f, df, G0, armijo=armijo, C1=C1, C2=C2, constC=constC, log=True, numItermaxEmd=numItermaxEmd, **kwargs)
        
        fgw_dist = log['loss'][-1]

        log['fgw_dist'] = fgw_dist
        log['u'] = log['u']
        log['v'] = log['v']
        return res, log
    else:
        if ot.__version__ < '0.9.0':
            return ot.gromov.cg(p, q, (1 - alpha) * M, alpha, f, df, G0, armijo=armijo, C1=C1, C2=C2, constC=constC, **kwargs)
        else:
            return ot.optim.cg(p, q, (1 - alpha) * M, alpha, f, df, G0, armijo=armijo, C1=C1, C2=C2, constC=constC, **kwargs)
        



import numpy as np
import scipy as s
import numpy.linalg as linalg
import scipy.stats as stats
import sys
from .basic_distributions import Distribution

from ..utils import *
# from mofapy2.core import gpu_utils
# TODO: Enable GPU support

class MultivariateGaussian(Distribution):
    """
    Class to define multivariate Gaussian distribution.
    This class can store:
    - if axis_cov=1 : N multivariate Gaussian Distributions of dimensionality D each
       (each row of a N x D matrix is a multivariate Gaussian)
    - if axis_cov=0 : D multivariate Gaussian Distributions of dimensionality N each
       (each column of a N x D X matrix is a multivariate Gaussian)
    
    For example, used with the Z node (N x K) this node can be used to store K multivariate Gaussians with dimensonality N each (axis_cov = 0).
    
    Equations (for axis_cov=0) :
    p(X|Mu,Sigma) = 1/(2pi)^{N/2} * 1/(|Sigma|^0.5) * exp( -0.5*(X-Mu)^{T} Sigma^{-1} (X-Mu) )
    log p(X|Mu,Sigma) = -N/2*log(2pi) - 0.5*log(|Sigma|) - 0.5*(X-Mu)^{T} Sigma^{-1} (X-Mu)
    E[X] = Mu
    E[X^2] = E[X]E[X] + Cov[X] = MuMu + Sigma
    cov[X] = Sigma
    H[X] = 0.5*log(|Sigma|) + N/2*(1+log(2*pi))

    Dimensionalities :
    - X: (N,D)
    - Mu: (N,D)
    - Sigma: (N,D,D) if axis_cov=1 or (D,N,N) if axis_cov=0
    - E[X]: (N,D)
    - E[X^2]: (N,D,D) if axis_cov=1 or (D,N,N) if axis_cov=0
    
    Parameters:
    - dim: (N,D)
    - mean: This can be an (N,D)-array, a scalar to be used as a constant mean vector for all Gaussians or
    a vector to be used for each multivariate Gaussian distirbution (Should have length D if axis_cov =1 and N otherwise)
    - cov: either a array of dimensions (N,N) for axis_cov=0 or (D,D) for axis_cov=1 that will be used as covariance matrix in each of the D or N multivariate Gaussian distirbutions
    or a list of length N (if axis_cov=1) or D (if axis_cov=0) or a tensor of dim (N,D,D) (if axis_cov=1) or (D,N,N) (if axis_cov=0)
    - axis_cov indicating which elementof dim (N,D) to use for dimenionality of the covariance matrix. If 0, it gives rise to D multivariate Gaussians with dimension N ( and (N,N)-covariance).
    If 1, it gives rise to N multivariate Gaussians with dimension D ( and (D,D)-covariance).
    """
    def __init__(self, dim, mean, cov, axis_cov=1, E=None):
        Distribution.__init__(self, dim)

        # Check dimensions are correct
        assert len(dim) == 2, "You need to specify two dimensions for 'dim': (number of distributions, dimensionality) "
        assert not (dim[0]==1 and dim[1]==1), "A 1x1 multivariate normal reduces to a Univariate normal "
        assert ((axis_cov == 0)or(axis_cov == 1)), "Error : axis_cov is the index of the dimension (N, D) for the covariance matrix, either 0 (N,N) or 1 (D,D)"

        ## Initialise the parameters ##

        ### Initialise the mean
        # If 'mean' is a scalar, broadcast it to all dimensions
        if isinstance(mean,(int,float)): mean = np.ones( (dim[0],dim[1]) ) * mean
        
        # If 'mean' has dim (D,) and we have N distributions, repeat it to all N distributions (for axis_cov=1) and correspondingly for axis_cov=0
        if len(mean.shape)==1 and mean.shape[0]==dim[1] and axis_cov==1: mean = mean * np.ones( (dim[0],dim[1]) ) 
        if len(mean.shape)==1 and mean.shape[0]==dim[0] and axis_cov==0: mean = (mean * np.ones( (dim[1],dim[0]) )).transpose()
        
        # check 'mean' has the right dimensions
        assert mean.shape[0]==dim[0] and mean.shape[1]==dim[1], "The given mean could not be broadcasted into a matrix with shape (N,D) "

        ### Initialise the covariance
        
        if isinstance(cov,np.ndarray):
            # If 'cov' is a matrix and not a tensor, broadcast it along the zeroth axis
            if len(cov.shape) == 2:
                if axis_cov == 1 :
                    assert cov.shape == (dim[1],dim[1]), "If providing a 2d-array, the covariance has to be of dim (D,D)"
                    cov = [cov] * dim[0]
                else:
                    assert cov.shape == (dim[0],dim[0]), "If providing a 2d-array, the covariance has to be of dim (N,N)"
                    cov = [cov] * dim[1]
                cov = np.array(cov)
        # If 'cov' is a list transform it to a tensor
        elif isinstance(cov, list):
            if axis_cov == 1:
                assert cov[0].shape ==  (dim[1],dim[1]) and len(cov) == dim[0], "If providing a list, the covariance has to be a list of length N with arrays of dim (D,D)"
            else:
                assert cov[0].shape ==  (dim[0],dim[0]) and len(cov) == dim[1], "If providing a list, the covariance has to be a list of length D with arrays of dim (N,N)"
            cov = np.array(cov)
        else: 
            print("The covariance needs to be a list or an array.")
            sys.exit()
            
        # check 'cov' has the right dimensions
        if axis_cov == 1:
            assert (cov.shape == (dim[0], dim[1], dim[1])), "The covariance could not be broadcasted into a tensor with shape (N,D,D)."
        if axis_cov == 0:
            assert (cov.shape == (dim[1], dim[0], dim[0])), "The covariance could not be broadcasted into a tensor with shape (D,N,N)."
            
        # store parameters
        self.axis_cov = axis_cov
        self.params = {'mean' : mean, 'cov' : cov }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E' : E }

    def updateExpectations(self):
        # Update first and second moments (of marginal only)
        E = self.params['mean']

        # second moment here of the marginal components: given by E(X_n^2) = E(X_n)^2 + Var(X_n)
        E2 = s.empty( (self.dim[0],self.dim[1]))
        if self.axis_cov == 1:
            for i in range(self.dim[0]):
               E2[i, :] = E[i,:]**2 + np.diag(self.params['cov'][i,:,:])
        else:
           for i in range(self.dim[1]):
               E2[:, i] = E[:,i]**2 + np.diag(self.params['cov'][i,:,:])

        self.expectations = {'E': E, 'E2': E2, 'cov' : self.params['cov']}

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        l = 0.

        if self.axis_cov == 1:
            D = self.dim[1]
            for n in range(self.dim[0]):
                qterm = (x[n,:]-self.params['mean'][n,:]).T.dot(linalg.det(self.params['cov'][n])).dot(x[n,:]-self.params['mean'][n,:])
                l += -0.5*D*np.log(2*np.pi) - 0.5*np.log(linalg.det(self.params['cov'][n])) - 0.5 * qterm
            # return np.sum( np.log(stats.multivariate_normal.pdf(x, mean=self.mean[n,:], cov=self.cov[n,:,:])) )

        else:
            N = self.dim[0]
            for d in range(self.dim[1]):
                qterm = (x[:, d] - self.params['mean'][:, d]).dot(linalg.det(self.params['cov'][d])).dot((x[:, d] - self.params['mean'][:, d]).T)
                l += -0.5 * N * np.log(2 * np.pi) - 0.5 * np.log(linalg.det(self.params['cov'][d])) - 0.5 * qterm

        return l

    def removeDimensions(self, axis, idx):
        # Method to remove undesired dimensions
        # - axis (int): axis from where to remove the elements
        # - idx (numpy array): indices of the elements to remove
        assert axis <= len(self.dim)
        assert np.all(idx < self.dim[axis])

        self.params["mean"] = np.delete(self.params["mean"], axis=axis, obj=idx)
        self.expectations["E"] = np.delete(self.expectations["E"], axis=axis, obj=idx)
        self.expectations["E2"] = np.delete(self.expectations["E2"], axis=axis, obj=idx)

        if self.axis_cov == 1: #cov has shape (N,D,D) for mean of shape (N,D)
            if axis == 0:
                self.params["cov"] = np.delete(self.params["cov"], axis=0, obj=idx)
            else:
                self.params["cov"] = np.delete(self.params["cov"], axis=1, obj=idx)
                self.params["cov"] = np.delete(self.params["cov"], axis=2, obj=idx)

        else: #cov has shape (D,N,N) for mean of shape (N,D)
            if axis == 0:
                self.params["cov"] = np.delete(self.params["cov"], axis=1, obj=idx)
                self.params["cov"] = np.delete(self.params["cov"], axis=2, obj=idx)
            else:
                self.params["cov"] = np.delete(self.params["cov"], axis=0, obj=idx)
        self.updateDim(axis=axis, new_dim=self.dim[axis] - len(idx))

    def sample(self):
        if axis_cov==1:
            samples = []
            for n in range(self.dim[0]):
                samples.append(np.random.multivariate_normal(self.params['mean'][n,:], self.params['cov'][n]))
            samples = np.array(samples)
        else:
            samples = []
            for d in range(self.dim[1]):
                samples.append(np.random.multivariate_normal(self.params['mean'][:,d], self.params['cov'][d]))
            samples = np.array(samples).T
        return samples

#    def entropy(self):
#         tmp = sum( [ logdet(self.cov[i,:,:]) for i in range(self.dim[0]) ] )
#         return ( 0.5*(tmp + (self.dim[0]*self.dim[1])*(1+np.log(2*pi)) ).sum() )


class MultivariateGaussian_reparam(Distribution):
    """
       Class to define multivariate Gaussian distribution with reparamtriation following Opper & Archambeau (2009)
\      Each distribution's mean and covariance are reparametrized as
       mu = K alpha
       Sigma = (K^(-1) + diag(lambda**2))^(-1)

       Dimensions:
       K : (D, N, N) if axis_cov = 0, else (N, D, D)
       alpha : (N, D)
       lambda: (N, D)
    """

    def __init__(self, dim, alpha, K, lamb, axis_cov=1, E =None):
        Distribution.__init__(self, dim)

        assert len(dim) == 2, "You need to specify two dimensions for 'dim': (number of distributions, dimensionality) "
        assert not (dim[0] == 1 and dim[1] == 1), "A 1x1 multivariate normal reduces to a Univariate normal "
        assert ((axis_cov == 0) or (
                    axis_cov == 1)), "Error : axis_cov is the index of the dimension (N, D) for the covariance matrix, either 0 (N,N) or 1 (D,D)"

        # Broadcast scalars to arrays for alpha and lambda
        if isinstance(alpha, (int, float)): alpha = np.ones((dim[0], dim[1])) * alpha
        if isinstance(lamb, (int, float)): lamb = np.ones((dim[0], dim[1])) * lamb

        # If 'alpha' has dim (D,) and we have N distributions, repeat it to all N distributions (for axis_cov=1) and correspondingly for axis_cov=0
        if len(alpha.shape) == 1 and alpha.shape[0] == dim[1] and axis_cov == 1: alpha = alpha * np.ones((dim[0], dim[1]))
        if len(alpha.shape) == 1 and alpha.shape[0] == dim[0] and axis_cov == 0: alpha = (
                    alpha * np.ones((dim[1], dim[0]))).transpose()
        if len(lamb.shape) == 1 and lamb.shape[0] == dim[1] and axis_cov == 1: lamb = lamb * np.ones(
            (dim[0], dim[1]))
        if len(lamb.shape) == 1 and lamb.shape[0] == dim[0] and axis_cov == 0: lamb = (
                lamb * np.ones((dim[1], dim[0]))).transpose()

        # Check dimensions of alpha and lamb are correct
        assert alpha.shape[0] == dim[0] and alpha.shape[1] == dim[1], "The given alpha could not be broadcasted into a matrix with shape (N,D) "
        assert lamb.shape[0] == dim[0] and lamb.shape[1] == dim[1], "The given lamb could not be broadcasted into a matrix with shape (N,D) "

        # check K has the right dimensions
        if isinstance(K, np.ndarray):
            # If 'K' is a matrix and not a tensor, broadcast it along the zeroth axis
            if len(K.shape) == 2:
                if axis_cov == 1:
                    assert K.shape == (
                    dim[1], dim[1]), "If providing a 2d-array, K has to be of dim (D,D)"
                    K = [K] * dim[0]
                else:
                    assert K.shape == (
                    dim[0], dim[0]), "If providing a 2d-array, K has to be of dim (N,N)"
                    K = [K] * dim[1]
                K = np.array(K)
        # If 'K' is a list transform it to a tensor
        elif isinstance(K, list):
            if axis_cov == 1:
                assert K[0].shape == (dim[1], dim[1]) and len(K) == dim[
                    0], "If providing a list, K has to be a list of length N with arrays of dim (D,D)"
            else:
                assert K[0].shape == (dim[0], dim[0]) and len(K) == dim[
                    1], "If providing a list, K has to be a list of length D with arrays of dim (N,N)"
            K = np.array(K)
        else:
            print("The input K needs to be a list or an array.")
            sys.exit()

        if axis_cov == 1:
            assert (K.shape == (
            dim[0], dim[1], dim[1])), "K could not be broadcasted into a tensor with shape (N,D,D)."
        if axis_cov == 0:
            assert (K.shape == (
            dim[1], dim[0], dim[0])), "K could not be broadcasted into a tensor with shape (D,N,N)."


        # Initialise the parameters
        self.axis_cov = axis_cov
        self.params = {'alpha': alpha, 'K': K, 'lamb' : lamb}

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E' : E }

    def updateExpectations(self):
        # Method to calculate expectation (N,D) and variance of th marginals (N,D)

        # first moments
        E = s.empty((self.dim[0], self.dim[1]))
        if self.axis_cov == 0:
            for i in range(self.dim[1]):
                E[:, i] = self.params['K'][i, :, :].dot(self.params['alpha'][:, i])
        elif self.axis_cov == 1:
            for i in range(self.dim[0]):
                E[i,:] = self.params['K'][i, :, :].dot(self.params['alpha'][i, :].transpose())

        # second moment here of the marginal components: given by E(X_n^2) = E(X_n)^2 + Var(X_n)
        E2 = s.empty((self.dim[0],self.dim[1]))
        if self.axis_cov == 0:
           for i in range(self.dim[1]):
               A = np.diag(self.params['lamb'][:, i]).dot(self.params['K'][i, :, :]).dot(
                   np.diag(self.params['lamb'][:, i])) + np.eye(self.dim[0])
               Ainv = np.linalg.inv(A)
               Sigma_diag = (1 / self.params['lamb'][:,i]**2) * (1 - np.diag(Ainv))
               E2[:, i] = E[:, i] ** 2 + Sigma_diag

        elif self.axis_cov == 1:
            for i in range(self.dim[0]):
                A = np.diag(self.params['lamb'][i, :]).dot(self.params['K']).dot(
                    np.diag(self.params['lamb'][i, :])) + np.eye(self.dim[1])
                Ainv = np.linalg.inv(A)
                Sigma_diag = (1 / self.params['lamb'][i,:]**2) * (1 - np.diag(Ainv))
                E2[i, :] = E[i, :] ** 2 + Sigma_diag

        self.expectations = {'E': E, 'E2': E2}

    def removeDimensions(self, axis, idx):
        # Method to remove undesired dimensions
        # - axis (int): axis from where to remove the elements
        # - idx (numpy array): indices of the elements to remove
        assert axis <= len(self.dim)
        assert np.all(idx < self.dim[axis])

        self.params["alpha"] = np.delete(self.params["alpha"], axis=axis, obj=idx)
        self.expectations["E"] = np.delete(self.expectations["E"], axis=axis, obj=idx)
        self.expectations["E2"] = np.delete(self.expectations["E2"], axis=axis, obj=idx)

        if self.axis_cov == 1: #K has shape (N,D,D) for mean of shape (N,D)
            if axis == 0:
                self.params["K"] = np.delete(self.params["K"], axis=0, obj=idx)
            else:
                self.params["K"] = np.delete(self.params["K"], axis=1, obj=idx)
                self.params["K"] = np.delete(self.params["K"], axis=2, obj=idx)

        else: #K has shape (D,N,N) for mean of shape (N,D)
            if axis == 0:
                self.params["K"] = np.delete(self.params["K"], axis=1, obj=idx)
                self.params["k"] = np.delete(self.params["K"], axis=2, obj=idx)
            else:
                self.params["K"] = np.delete(self.params["K"], axis=0, obj=idx)
        self.updateDim(axis=axis, new_dim=self.dim[axis] - len(idx))



import scanpy as sc
import os
import torch
import time
import torch as th
import torch.distributed as dist
import numpy as np
class scDiffusion(object):

    def __init__(
            self,
            adata,
            device='cuda:0'
    ):
        from ..utils import check_dependencies
        check_dependencies(
            ['blobfile','mpi4py']
        )
        self.adata=adata
        self.device=device
        self.num_genes=self.adata.shape[1]

    def prepare_VAE(
            self,
            state_dict=None,
            cell_type='celltype',
            batch_size=128,
            num_genes=18996,
            seed=1234,
            loss_ae="mse",
            decoder_activation='ReLU',
    ):
        from ..externel.scdiffusion.VAE.VAE_model import VAE
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data,prepare_data
        print('device: ',self.device)

        datasets = prepare_data(
            adata=self.adata,
            batch_size=batch_size,
            train_vae=True,
            cell_type=cell_type,
        )

        autoencoder = VAE(
            num_genes=num_genes,
            device=self.device,
            seed=seed,
            loss_ae=loss_ae,
            hidden_dim=128,
            decoder_activation=decoder_activation,
        )
        if state_dict is not None:
            print('loading pretrained model from: \n',state_dict)
            use_gpu = "cuda" in self.device
            autoencoder.encoder.load_state(state_dict["encoder"], use_gpu)
            autoencoder.decoder.load_state(state_dict["decoder"], use_gpu)

        return autoencoder, datasets



    def train_VAE(
            self,
            state_dict=None,
            cell_type='celltype',
            loss_ae="mse",
            decoder_activation='ReLU',
            local_rank=0,
            split_seed=1234,
            num_genes=18996,
            seed=1234,
            hparams="",
            batch_size=128,
            max_steps=200000,
            max_minutes=3000,
            checkpoint_freq=50000,
            save_dir=None,
            sweep_seeds=200,
            return_model=False,
    ):
        
        if save_dir is None:
            save_dir = os.path.join(os.getcwd(), "result/VAE_model")
            os.makedirs(save_dir, exist_ok=True)
        
        if state_dict is not None:
            filenames = {}
            checkpoint_path = {
                "encoder": os.path.join(
                    state_dict, filenames.get("model", "encoder.ckpt")
                ),
                "decoder": os.path.join(
                    state_dict, filenames.get("model", "decoder.ckpt")
                ),
                "gene_order": os.path.join(
                    state_dict, filenames.get("gene_order", "gene_order.tsv")
                ),
            }
            autoencoder, datasets = self.prepare_VAE(
                state_dict=checkpoint_path,
                cell_type=cell_type,
                batch_size=batch_size,
                num_genes=num_genes,
                seed=seed,
                loss_ae=loss_ae,
                decoder_activation=decoder_activation)
        else:
            autoencoder, datasets = self.prepare_VAE(
                state_dict=None,
                cell_type=cell_type,
                batch_size=batch_size,
                num_genes=num_genes,
                seed=seed,
                loss_ae=loss_ae,
                decoder_activation=decoder_activation)
        hparams = autoencoder.hparams

        start_time = time.time()
        from tqdm import tqdm
        for step in tqdm(range(max_steps)):

            genes, _ = next(datasets)

            minibatch_training_stats = autoencoder.train(genes)

            if step % 1000 == 0:
                for key, val in minibatch_training_stats.items():
                    print('step ', step, 'loss ', val)

            ellapsed_minutes = (time.time() - start_time) / 60

            stop = ellapsed_minutes > max_minutes or (
                step == max_steps - 1
            )

            if ((step % checkpoint_freq) == 0 or stop):

                os.makedirs(save_dir,exist_ok=True)
                torch.save(
                    autoencoder.state_dict(),
                    os.path.join(
                        save_dir,
                        "model_seed={}_step={}.pt".format(seed, step),
                    ),
                )

                if stop:
                    break

        if return_model:
            return autoencoder, datasets
        

    def train_cellmodel(
            self,
            cell_type='celltype',
            schedule_sampler="uniform",
            lr=1e-4,
            weight_decay=0.0001,
            lr_anneal_steps=500000,
            batch_size=12,
            microbatch=-1,  # -1 disables microbatches
            ema_rate="0.9999",  # comma-separated list of EMA values
            log_interval=100,
            save_interval=200000,
            resume_checkpoint="",
            use_fp16=False,
            fp16_scale_growth=1e-3,
            vae_path = 'output/Autoencoder_checkpoint/muris_AE/model_seed=0_step=0.pt',
            model_name="muris_diffusion",
            save_dir='output/diffusion_checkpoint'
    ):
        from ..externel.scdiffusion.guided_diffusion import dist_util, logger
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data
        from ..externel.scdiffusion.guided_diffusion.resample import create_named_schedule_sampler
        from ..externel.scdiffusion.guided_diffusion.script_util import (
            model_and_diffusion_defaults,
            create_model_and_diffusion,
            args_to_dict,
            add_dict_to_argparser,
        )
        from ..externel.scdiffusion.guided_diffusion.train_util import TrainLoop
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data,prepare_data

        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        dist_util.setup_dist()
        logger.configure(dir='../output/logs/'+model_name)  # log file

        logger.log("creating model and diffusion...")
        model, diffusion = create_model_and_diffusion(
                input_dim = 128,
                hidden_dim = [512,512,256,128],
                dropout = 0.0,
                learn_sigma=False,
                diffusion_steps=1000,
                noise_schedule="linear",
                timestep_respacing="",
                use_kl=False,
                predict_xstart=False,
                rescale_timesteps=False,
                rescale_learned_sigmas=False,
                class_cond=False,
        )
        model.to(dist_util.dev())
        schedule_sampler = create_named_schedule_sampler(schedule_sampler, diffusion)

        logger.log("creating data loader...")
        data = prepare_data(
            adata=self.adata, 
            batch_size=batch_size,
            vae_path=vae_path,
            train_vae=False,
            cell_type=cell_type,
        )


        logger.log("training...")
        TrainLoop(
            model=model,
            diffusion=diffusion,
            data=data,
            batch_size=batch_size,
            microbatch=microbatch,
            lr=lr,
            ema_rate=ema_rate,
            log_interval=log_interval,
            save_interval=save_interval,
            resume_checkpoint=resume_checkpoint,
            use_fp16=use_fp16,
            fp16_scale_growth=fp16_scale_growth,
            schedule_sampler=schedule_sampler,
            weight_decay=weight_decay,
            lr_anneal_steps=lr_anneal_steps,
            model_name=model_name,
            save_dir=save_dir
        ).run_loop()


    def train_classifier(
            self,
            model_name="classifier_muris",
            save_dir='output/classifier_checkpoint/classifier_muris',
            noised=True,
            resume_checkpoint="",
            use_fp16=False,
            batch_size=128,
            vae_path='output/Autoencoder_checkpoint/muris_AE/model_seed=0_step=0.pt',
            cell_type='celltype',
            val_adata=None,
            input_dim=128,
            lr=3e-4,
            weight_decay=0.0,
            start_guide_time=500,
            microbatch=-1,
            model_path='output/classifier_checkpoint/classifier_muris',
            iterations=500000,
            log_interval=100,
            eval_interval=100,
            save_interval=100000,
            anneal_lr=False,
            num_class=12,
            schedule_sampler="uniform",
            
    ):
        from ..externel.scdiffusion.guided_diffusion import dist_util, logger
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data
        from ..externel.scdiffusion.guided_diffusion.resample import create_named_schedule_sampler
        from ..externel.scdiffusion.guided_diffusion.script_util import (
            model_and_diffusion_defaults,
            create_model_and_diffusion,
            args_to_dict,
            add_dict_to_argparser,
            create_classifier_and_diffusion
        )
        from ..externel.scdiffusion.guided_diffusion.train_util import log_loss_dict,parse_resume_step_from_filename
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data,prepare_data
        from torch.nn.parallel.distributed import DistributedDataParallel as DDP
        from torch.optim import AdamW
        import torch.nn.functional as F

        from ..externel.scdiffusion.guided_diffusion.fp16_util import MixedPrecisionTrainer

        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        dist_util.setup_dist()
        logger.configure(dir='../output/logs/'+model_name)  # log file

        logger.log("creating model and diffusion...")
        model, diffusion = create_classifier_and_diffusion(
                input_dim = input_dim,
                hidden_dim = [512,512,256,128],
                dropout = 0.1,
                classifier_use_fp16=use_fp16,
                num_class = num_class,
                learn_sigma=False,
                diffusion_steps=1000,
                noise_schedule="linear",
                timestep_respacing="",
                use_kl=False,
                predict_xstart=False,
                rescale_timesteps=False,
                rescale_learned_sigmas=False,
                class_cond=False,
        )
        model.to(dist_util.dev())
        if noised:
            schedule_sampler = create_named_schedule_sampler(
                schedule_sampler, diffusion
            )
        resume_step = 0
        if resume_checkpoint:
            resume_step = parse_resume_step_from_filename(resume_checkpoint)
            if dist.get_rank() == 0:
                logger.log(
                    f"loading model from checkpoint: {resume_checkpoint}... at {resume_step} step"
                )
                model.load_state_dict(
                    dist_util.load_state_dict(
                        resume_checkpoint, map_location=dist_util.dev()
                    )
                )

        # Needed for creating correct EMAs and fp16 parameters.
        dist_util.sync_params(model.parameters())

        mp_trainer = MixedPrecisionTrainer(
            model=model, use_fp16=use_fp16, initial_lg_loss_scale=16.0
        )

        model = DDP(
            model,
            device_ids=[dist_util.dev()],
            output_device=dist_util.dev(),
            broadcast_buffers=False,
            bucket_cap_mb=128,
            find_unused_parameters=True,
        )

        logger.log("creating data loader...")
        data = prepare_data(
            adata=self.adata, 
            batch_size=batch_size,
            vae_path=vae_path,
            train_vae=False,
            cell_type=cell_type,
        )
        if val_adata is not None:
            val_data = prepare_data(
                adata=val_adata, 
                batch_size=batch_size,
                vae_path=vae_path,
                hidden_dim=128,
                train_vae=False,
                cell_type=cell_type
            )
        else:
            val_data = None
        logger.log(f"creating optimizer...")
        opt = AdamW(mp_trainer.master_params, lr=lr, weight_decay=weight_decay)
        if resume_checkpoint:
            opt_checkpoint = os.path.join(
                os.path.dirname(resume_checkpoint), f"opt{resume_step:06}.pt"
            )
            logger.log(f"loading optimizer state from checkpoint: {opt_checkpoint}")
            opt.load_state_dict(
                dist_util.load_state_dict(opt_checkpoint, map_location=dist_util.dev())
            )

        logger.log("training classifier model...")

        def forward_backward_log(data_loader, prefix="train"):
            batch, extra = next(data_loader)
            labels = extra["y"].to(dist_util.dev())

            batch = batch.to(dist_util.dev())
            # Noisy cells
            if noised:
                t, _ = schedule_sampler.sample(batch.shape[0], dist_util.dev(), start_guide_time=start_guide_time)
                batch = diffusion.q_sample(batch, t)
            else:
                t = th.zeros(batch.shape[0], dtype=th.long, device=dist_util.dev())

            for i, (sub_batch, sub_labels, sub_t) in enumerate(
                split_microbatches(microbatch, batch, labels, t)
            ):
                logits = model(sub_batch, sub_t)
                loss = F.cross_entropy(logits, sub_labels, reduction="none")

                losses = {}
                losses[f"{prefix}_loss"] = loss.detach()
                losses[f"{prefix}_acc@1"] = compute_top_k(
                    logits, sub_labels, k=1, reduction="none"
                )

                log_loss_dict(diffusion, sub_t, losses)
                del losses
                loss = loss.mean()
                if loss.requires_grad:
                    if i == 0:
                        mp_trainer.zero_grad()
                    mp_trainer.backward(loss * len(sub_batch) / len(batch))
        model_path = save_dir
        for step in range(iterations - resume_step):
            logger.logkv("step", step + resume_step)
            logger.logkv(
                "samples",
                (step + resume_step + 1) * batch_size * dist.get_world_size(),
            )
            if anneal_lr:
                set_annealed_lr(opt, lr, (step + resume_step) / iterations)
            forward_backward_log(data)
            mp_trainer.optimize(opt)
            if val_data is not None and not step % eval_interval:
                with th.no_grad():
                    with model.no_sync():
                        model.eval()
                        forward_backward_log(val_data, prefix="val")
                        model.train()
            if not step % log_interval:
                logger.dumpkvs()
            if (
                step
                and dist.get_rank() == 0
                and not (step + resume_step) % save_interval
            ):
                logger.log("saving model...")
                save_model(mp_trainer, opt, step + resume_step, model_path)

        if dist.get_rank() == 0:
            logger.log("saving model...")
            save_model(mp_trainer, opt, step + resume_step, model_path)
        dist.barrier()




    def load_VAE(self,
            vae_path='output/checkpoint/AE/open_problem/model_seed=1234_step=150000.pt',
            loss_ae="mse",
            decoder_activation='ReLU',
    ):
        from ..externel.scdiffusion.VAE.VAE_model import VAE
        autoencoder = VAE(
            num_genes=self.num_genes,
            device=self.device,
            seed=0,
            loss_ae=loss_ae,
            hidden_dim=128,
            decoder_activation=decoder_activation,
        )
        autoencoder.load_state_dict(torch.load(vae_path))
        return autoencoder

    def generate_cell(
            self,
            input_dim=128,
            clip_denoised=False,
            num_samples=3000,
            batch_size=3000,
            use_ddim=False,
            model_path="output/diffusion_checkpoint/muris_diffusion/model600000.pt",
            vae_path='output/Autoencoder_checkpoint/muris_AE/model_seed=0_step=0.pt',
            sample_dir="output/simulated_samples/muris",
            loss_ae="mse",
            decoder_activation='ReLU',
    ):
        from ..externel.scdiffusion.guided_diffusion import dist_util, logger
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data
        from ..externel.scdiffusion.guided_diffusion.resample import create_named_schedule_sampler
        from ..externel.scdiffusion.guided_diffusion.script_util import (
            model_and_diffusion_defaults,
            create_model_and_diffusion,
            args_to_dict,
            add_dict_to_argparser,
        )
        from ..externel.scdiffusion.guided_diffusion.train_util import TrainLoop
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data,prepare_data

        dist_util.setup_dist()
        logger.configure(dir='checkpoint/sample_logs')

        logger.log("creating model and diffusion...")
        model, diffusion = create_model_and_diffusion(
                input_dim = input_dim,
                hidden_dim = [512,512,256,128],
                dropout = 0.0,
                learn_sigma=False,
                diffusion_steps=1000,
                noise_schedule="linear",
                timestep_respacing="",
                use_kl=False,
                predict_xstart=False,
                rescale_timesteps=False,
                rescale_learned_sigmas=False,
                class_cond=False,
        )
        model.load_state_dict(
            dist_util.load_state_dict(model_path, map_location="cpu")
        )
        model.to(dist_util.dev())
        model.eval()

        logger.log("sampling...")
        all_cells = []
        while len(all_cells) * batch_size < num_samples:
            model_kwargs = {}
            sample_fn = (
                diffusion.p_sample_loop if not use_ddim else diffusion.ddim_sample_loop
            )
            sample, traj = sample_fn(
                model,
                (batch_size, input_dim), 
                clip_denoised=clip_denoised,
                model_kwargs=model_kwargs,
                start_time=diffusion.betas.shape[0],
            )

            gathered_samples = [th.zeros_like(sample) for _ in range(dist.get_world_size())]
            dist.all_gather(gathered_samples, sample)  # gather not supported with NCCL
            all_cells.extend([sample.cpu().numpy() for sample in gathered_samples])
            logger.log(f"created {len(all_cells) * batch_size} samples")

        arr = np.concatenate(all_cells, axis=0)
        dist.barrier()
        autoencoder = self.load_VAE(
            vae_path=vae_path,
            loss_ae=loss_ae,
            decoder_activation=decoder_activation,
        )
        cell_gen = autoencoder(torch.tensor(arr).cuda(),return_decoded=True).cpu().detach().numpy()

        adata=sc.AnnData(cell_gen)
        adata.var.index=self.adata.var_names
        logger.log("sampling complete")
        return adata

        
        

    def generate_classifier(
            self,
            cell_type=[0], multi=False, inter=False, weight=[10,10],
            input_dim=128,
            hidden_dim = [512,512,256,128],
            dropout=0.0,
            clip_denoised=True,
            num_samples=9000,
            batch_size=3000,
            use_ddim=False,
            class_cond=False, 
            loss_ae="mse",
            decoder_activation='ReLU',

            model_path="output/diffusion_checkpoint/muris_diffusion/model000000.pt", 

            # ***if commen conditional generation & gradiante interpolation, use this path***
            classifier_path="output/classifier_checkpoint/classifier_muris/model000100.pt",
            # ***if multi-conditional, use this path. replace this to your own classifiers***
            classifier_path1="output/classifier_checkpoint/classifier_muris_ood_type/model200000.pt",
            classifier_path2="output/classifier_checkpoint/classifier_muris_ood_organ/model200000.pt",
            num_class1 = 2,  # set this to the number of classes in your own dataset. this is the first condition (for example cell organ).
            num_class2 = 2,  # this is the second condition (for example cell type).

            # ***if commen conditional generation, use this scale***
            classifier_scale=2,
            # ***in multi-conditional, use this scale. scale1 and scale2 are the weights of two classifiers***
            # ***in Gradient Interpolation, use this scale, too. scale1 and scale2 are the weights of two gradients***
            

            # ***if gradient interpolation, replace these base on your own situation***
            vae_path='output/Autoencoder_checkpoint/WOT/model_seed=0_step=150000.pt', 
            num_gene=19423,
            num_class=12,
            init_time = 600,    # initial noised state if interpolation
            init_cell_adata = None,   #input initial noised cell state

            sample_dir=f"output/simulated_samples/muris",
            start_guide_steps = 500,     # the time to use classifier guidance
            filter = False,   # filter the simulated cells that are classified into other condition, might take long time

    ):
        classifier_scale1=weight[0]*2/10,
        classifier_scale2=weight[1]*2/10,
        import torch.nn.functional as F
        from ..externel.scdiffusion.guided_diffusion import dist_util, logger
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data
        from ..externel.scdiffusion.guided_diffusion.resample import create_named_schedule_sampler
        from ..externel.scdiffusion.guided_diffusion.script_util import (
            NUM_CLASSES,
            model_and_diffusion_defaults,
            classifier_and_diffusion_defaults,
            create_model_and_diffusion,
            create_classifier,
            add_dict_to_argparser,
            args_to_dict,
        )
        from ..externel.scdiffusion.guided_diffusion.train_util import TrainLoop
        from ..externel.scdiffusion.guided_diffusion.cell_datasets_loader import load_data,prepare_data

        dist_util.setup_dist()
        logger.configure(dir='checkpoint/sample_logs')

        logger.log("creating model and diffusion...")
        model, diffusion = create_model_and_diffusion(
                input_dim = input_dim,
                hidden_dim = [512,512,256,128],
                dropout = 0.0,
                learn_sigma=False,
                diffusion_steps=1000,
                noise_schedule="linear",
                timestep_respacing="",
                use_kl=False,
                predict_xstart=False,
                rescale_timesteps=False,
                rescale_learned_sigmas=False,
                class_cond=False,
        )
        model.load_state_dict(
            dist_util.load_state_dict(model_path, map_location="cpu")
        )
        model.to(dist_util.dev())
        model.eval()

        logger.log("loading classifier...")
        if multi:
            num_class = num_class1 # how many classes in this condition
            classifier1 = create_classifier(
                input_dim=input_dim,
                hidden_dim=hidden_dim,
                num_class=num_class1,
                dropout=dropout,
            )
            classifier1.load_state_dict(
                dist_util.load_state_dict(classifier_path1, map_location="cpu")
            )
            classifier1.to(dist_util.dev())
            classifier1.eval()

            num_class = num_class2 # how many classes in this condition
            classifier2 = create_classifier(
                input_dim=input_dim,
                hidden_dim=hidden_dim,
                num_class=num_class2,
                dropout=dropout,
            )
            classifier2.load_state_dict(
                dist_util.load_state_dict(classifier_path2, map_location="cpu")
            )
            classifier2.to(dist_util.dev())
            classifier2.eval()

        else:
            classifier = create_classifier(
                input_dim=input_dim,
                hidden_dim=hidden_dim,
                num_class=num_class,
                dropout=dropout,
            )
            classifier.load_state_dict(
                dist_util.load_state_dict(classifier_path, map_location="cpu")
            )
            classifier.to(dist_util.dev())
            classifier.eval()

        '''
        control function for Gradient Interpolation Strategy
        '''
        def cond_fn_inter(x, t, y=None, init=None, diffusion=None):
            assert y is not None
            y1 = y[:,0]
            y2 = y[:,1]
            # xt = diffusion.q_sample(th.tensor(init,device=dist_util.dev()),t*th.ones(init.shape[0],device=dist_util.dev(),dtype=torch.long),)
            with th.enable_grad():
                x_in = x.detach().requires_grad_(True)
                logits = classifier(x_in, t)
                log_probs = F.log_softmax(logits, dim=-1)
                selected1 = log_probs[range(len(logits)), y1.view(-1)]
                selected2 = log_probs[range(len(logits)), y2.view(-1)]
                
                grad1 = th.autograd.grad(selected1.sum(), x_in, retain_graph=True)[0] * classifier_scale1
                grad2 = th.autograd.grad(selected2.sum(), x_in, retain_graph=True)[0] * classifier_scale2

                # l2_loss = ((x_in-xt)**2).mean()
                # grad3 = th.autograd.grad(-l2_loss, x_in, retain_graph=True)[0] * 100

                return grad1+grad2#+grad3

        '''
        control function for multi-conditional generation
        Two conditional generation here
        '''
        def cond_fn_multi(x, t, y=None):
            assert y is not None
            y1 = y[:,0]
            y2 = y[:,1]
            with th.enable_grad():
                x_in = x.detach().requires_grad_(True)
                logits1 = classifier1(x_in, t)
                log_probs1 = F.log_softmax(logits1, dim=-1)
                selected1 = log_probs1[range(len(logits1)), y1.view(-1)]

                logits2 = classifier2(x_in, t)
                log_probs2 = F.log_softmax(logits2, dim=-1)
                selected2 = log_probs2[range(len(logits2)), y2.view(-1)]
                
                grad1 = th.autograd.grad(selected1.sum(), x_in, retain_graph=True)[0] * classifier_scale1
                grad2 = th.autograd.grad(selected2.sum(), x_in, retain_graph=True)[0] * classifier_scale2
                
                return grad1+grad2

        '''
        control function for one conditional generation
        '''
        def cond_fn_ori(x, t, y=None):
            assert y is not None
            with th.enable_grad():
                x_in = x.detach().requires_grad_(True)
                logits = classifier(x_in, t)
                log_probs = F.log_softmax(logits, dim=-1)
                selected = log_probs[range(len(logits)), y.view(-1)]
                grad = th.autograd.grad(selected.sum(), x_in, retain_graph=True)[0] * classifier_scale
                return grad
            
        def model_fn(x, t, y=None, init=None, diffusion=None):
            assert y is not None
            if class_cond:
                return model(x, t, y if class_cond else None)
            else:
                return model(x, t)

        if inter:
            # input real cell expression data as initial noise
            ori_adata = init_cell_adata

        logger.log("sampling...")
        all_cell = []
        sample_num = 0
        while sample_num < num_samples:
            model_kwargs = {}

            if not multi and not inter:
                classes = (cell_type[0])*th.ones((batch_size,), device=dist_util.dev(), dtype=th.long)

            if multi:
                classes1 = (cell_type[0])*th.ones((batch_size,), device=dist_util.dev(), dtype=th.long)
                classes2 = (cell_type[1])*th.ones((batch_size,), device=dist_util.dev(), dtype=th.long)
                # classes3 = ... if more conditions
                classes = th.stack((classes1,classes2), dim=1)

            if inter:
                classes1 = (cell_type[0])*th.ones((batch_size,), device=dist_util.dev(), dtype=th.long)
                classes2 = (cell_type[1])*th.ones((batch_size,), device=dist_util.dev(), dtype=th.long)
                classes = th.stack((classes1,classes2), dim=1)

            model_kwargs["y"] = classes
            sample_fn = (
                diffusion.p_sample_loop if not use_ddim else diffusion.ddim_sample_loop
            )

            if inter:
                celltype = ori_adata.obs['period'].cat.categories.tolist()[cell_type[0]]
                adata = ori_adata[ori_adata.obs['period']==celltype].copy()

                start_x = adata.X
                autoencoder = self.load_VAE(vae_path, self.num_genes)
                start_x = autoencoder(torch.tensor(start_x,device=dist_util.dev()),return_latent=True).detach().cpu().numpy()

                n, m = start_x.shape  
                if n >= batch_size:  
                    start_x = start_x[:batch_size, :]  
                else:  
                    repeat_times = batch_size // n  
                    remainder = batch_size % n  
                    start_x = np.concatenate([start_x] * repeat_times + [start_x[:remainder, :]], axis=0)  
                
                noise = diffusion.q_sample(th.tensor(start_x,device=dist_util.dev()),
                            init_time*th.ones(start_x.shape[0],device=dist_util.dev(),dtype=torch.long),)
                model_kwargs["init"] = start_x
                model_kwargs["diffusion"] = diffusion

            if multi:
                sample, traj = sample_fn(
                    model_fn,
                    (batch_size, input_dim),
                    clip_denoised=clip_denoised,
                    model_kwargs=model_kwargs,
                    cond_fn=cond_fn_multi,
                    device=dist_util.dev(),
                    noise = None,
                    start_time=diffusion.betas.shape[0],
                    start_guide_steps=start_guide_steps,
                )
            elif inter:
                sample, traj = sample_fn(
                    model_fn,
                    (batch_size, input_dim),
                    clip_denoised=clip_denoised,
                    model_kwargs=model_kwargs,
                    cond_fn=cond_fn_inter,
                    device=dist_util.dev(),
                    noise = noise,
                    start_time=diffusion.betas.shape[0],
                    start_guide_steps=start_guide_steps,
                )
            else:
                sample, traj = sample_fn(
                    model_fn,
                    (batch_size, input_dim),
                    clip_denoised=clip_denoised,
                    model_kwargs=model_kwargs,
                    cond_fn=cond_fn_ori,
                    device=dist_util.dev(),
                    noise = None,
                )

            gathered_samples = [th.zeros_like(sample) for _ in range(dist.get_world_size())]
            dist.all_gather(gathered_samples, sample)  # gather not supported with NCCL
            if filter:
                for sample in gathered_samples:
                    if multi:
                        logits1 = classifier1(sample, torch.zeros((sample.shape[0]), device=sample.device))
                        logits2 = classifier2(sample, torch.zeros((sample.shape[0]), device=sample.device))
                        prob1 = F.softmax(logits1, dim=-1)
                        prob2 = F.softmax(logits2, dim=-1)
                        type1 = torch.argmax(prob1, 1)
                        type2 = torch.argmax(prob2, 1)
                        select_index = ((type1 == cell_type[0]) & (type2 == cell_type[1]))
                        all_cell.extend([sample[select_index].cpu().numpy()])
                        sample_num += select_index.sum().item()
                    elif inter:
                        logits = classifier(sample, torch.zeros((sample.shape[0]), device=sample.device))
                        prob = F.softmax(logits, dim=-1)
                        left = (prob[:,cell_type[0]] > weight[0]/10-0.15) & (prob[:,cell_type[0]] < weight[0]/10+0.15)
                        right = (prob[:,cell_type[1]] > weight[1]/10-0.15) & (prob[:,cell_type[1]] < weight[1]/10+0.15)
                        select_index = left & right
                        all_cell.extend([sample[select_index].cpu().numpy()])
                        sample_num += select_index.sum().item()
                    else:
                        logits = classifier(sample, torch.zeros((sample.shape[0]), device=sample.device))
                        prob = F.softmax(logits, dim=-1)
                        type = torch.argmax(prob, 1)
                        select_index = (type == cell_type[0])
                        all_cell.extend([sample[select_index].cpu().numpy()])
                        sample_num += select_index.sum().item()
                logger.log(f"created {sample_num} samples")
            else:
                all_cell.extend([sample.cpu().numpy() for sample in gathered_samples])
                sample_num = len(all_cell) * batch_size
                logger.log(f"created {len(all_cell) * batch_size} samples")

        arr = np.concatenate(all_cell, axis=0)
        dist.barrier()
        autoencoder = self.load_VAE(
            vae_path=vae_path,
            loss_ae=loss_ae,
            decoder_activation=decoder_activation,
        )
        cell_gen = autoencoder(torch.tensor(arr).cuda(),return_decoded=True).cpu().detach().numpy()

        adata=sc.AnnData(cell_gen)
        adata.var.index=self.adata.var_names
        logger.log("sampling complete")
        return adata
            


def split_microbatches(microbatch, *args):
    bs = len(args[0])
    if microbatch == -1 or microbatch >= bs:
        yield tuple(args)
    else:
        for i in range(0, bs, microbatch):
            yield tuple(x[i : i + microbatch] if x is not None else None for x in args)

def compute_top_k(logits, labels, k, reduction="mean"):
    _, top_ks = th.topk(logits, k, dim=-1)
    if reduction == "mean":
        return (top_ks == labels[:, None]).float().sum(dim=-1).mean().item()
    elif reduction == "none":
        return (top_ks == labels[:, None]).float().sum(dim=-1)

def set_annealed_lr(opt, base_lr, frac_done):
    lr = base_lr * (1 - frac_done)
    for param_group in opt.param_groups:
        param_group["lr"] = lr

def save_model(mp_trainer, opt, step, model_path):
    if dist.get_rank() == 0:
        model_dir = model_path
        os.makedirs(model_dir,exist_ok=True)
        th.save(
            mp_trainer.master_params_to_state_dict(mp_trainer.master_params),
            os.path.join(model_dir, f"model{step:06d}.pt"),
        )
        th.save(opt.state_dict(), os.path.join(model_dir, f"opt{step:06d}.pt"))
